# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

from PyQt4.QtGui import *
from PyQt4.QtCore import *
import time, types, sys, re
from operator import attrgetter
import anki, anki.utils, ankiqt.forms
from ankiqt import ui
from anki.utils import fmtTimeSpan, parseTags, addTags, deleteTags
from anki.errors import *

# Deck editor
##########################################################################

class DeckModel(QAbstractTableModel):

    def __init__(self, deck):
        QAbstractTableModel.__init__(self)
        self.origDeck = deck
        self.filterTag = None
        self.deleted = set()
        # column title, display accessor, sort attr
        self.columns = [(_("Question"), self.currentQuestion,
                         self.currentQuestion),
                        (_("Answer"), self.currentAnswer,
                         self.currentAnswer),
                        (_("Next due"), self.nextDue,
                         "nextTime")]
        self.showMatching()

    # Model interface
    ######################################################################

    def rowCount(self, index):
        return len(self.deck)

    def columnCount(self, index):
        return len(self.columns)

    def data(self, index, role):
        if not index.isValid():
            return QVariant()
        if role == Qt.FontRole and index.column() == 2:
            f = QFont()
            f.setPixelSize(12)
            return QVariant(f)
        elif role == Qt.DisplayRole or role == Qt.EditRole:
            s = self.columns[index.column()][1](self.deck[index.row()])
            s = s.replace("<br>", u" ")
            s = s.replace("\n", u"  ")
            s = ui.utils.stripHTML(s)
            s = s.strip()
            return QVariant(s)
        elif role == Qt.SizeHintRole:
            if index.column() == 2:
                return QVariant(20)
        else:
            return QVariant()

    def headerData(self, section, orientation, role):
        if orientation == Qt.Vertical:
            return QVariant()
        elif role == Qt.DisplayRole:
            return QVariant(self.columns[section][0])
        else:
            return QVariant()

    def flags(self, index):
        return Qt.ItemFlag(Qt.ItemIsEnabled |
                           Qt.ItemIsSelectable)

    # Filtering
    ######################################################################

    def showMatching(self, searchStr=None, tag=None):
        if not searchStr and not tag:
            self.deck = list(self.origDeck)
        else:
            self.deck = [card for card in self.origDeck
                         if (self.cardContainsString(card, searchStr) and
                             self.cardIsTagged(card, tag))]
        try:
            self.sort(*self.currentSortScheme)
        except AttributeError:
            # not inited yet - no problem
            pass

    def cardContainsString(self, card, searchStr):
        if not searchStr:
            return True
        str = (card.question + card.answer).lower()
        str = re.sub("</?.*?>", "", str)
        return searchStr.lower() in str

    def cardIsTagged(self, card, tag):
        if not tag:
            return True
        return card.hasTag(tag)

    # Sorting
    ######################################################################

    def sort(self, kind, value, reverse=False):
        if kind == "attr":
            func = attrgetter(value)
            self.deck.sort(key=func, reverse=reverse)
        elif kind == "field":
            self.deck.sort(cmp=lambda a,b,f=value: self.cmpFields(a,b,f),
                           reverse=reverse)
        elif kind == "special":
            if value == "ease":
                self.deck.sort(cmp=self.cmpEase, reverse=reverse)
        self.currentSortScheme = (kind, value, reverse)
        self.reset()

    def cmpFields(self, a, b, field):
        ans = cmp(a.model, b.model)
        if ans:
            return ans
        aye = a.fact.get(field, "")
        bee = b.fact.get(field, "")
        try:
            return cmp(float(aye), float(bee))
        except ValueError:
            return cmp(aye, bee)

    def cmpEase(self, a, b):
        astats = self.origDeck.sched.summarizeStats(a.stats)
        bstats = self.origDeck.sched.summarizeStats(b.stats)
        try:
            aratio = astats['YesTotal'] / float(astats['Total'])
        except ZeroDivisionError:
            aratio = -1
        try:
            bratio = bstats['YesTotal'] / float(bstats['Total'])
        except ZeroDivisionError:
            bratio = -1
        return cmp(aratio, bratio)

    # Tools
    ######################################################################

    def getCard(self, index):
        return self.deck[index.row()]

    def isDeleted(self, card):
        return card in self.deleted

    def isChanged(self, card):
        return False
        #return card in self.changed

    def cardIndex(self, card):
        "Return the index of CARD, if currently displayed."
        return self.deck.index(card)

    # Card accessors
    ######################################################################

    def currentQuestion(self, card):
        return card.question

    def currentAnswer(self, card):
        return card.answer

    def nextDue(self, card):
        secs = card.nextTime - time.time()
        if secs <= 0:
            if self.origDeck.sched.cardIsNew(card):
                return _("(new card)")
            else:
                return _("%s ago") % fmtTimeSpan(abs(secs), pad=0)
        else:
            return _("in %s") % fmtTimeSpan(secs, pad=0)

    def created(self, card):
        fmt = "%y/%m/%d %H:%M"
        return time.strftime(fmt, time.localtime(card.created))

class StatusDelegate(QItemDelegate):

    def __init__(self, parent, model):
        QItemDelegate.__init__(self, parent)
        self.model = model

    def paint(self, painter, option, index):
        card = self.model.getCard(index)
        standard = True
        # tagged
        if self.model.origDeck.sched.cardIsSuspended(card):
            brush = QBrush(QColor("#ffffaa"))
            standard = False
        if card.hasTag("Marked"):
            brush = QBrush(QColor("#aaaaff"))
            standard = False
        if self.model.isDeleted(card):
            brush = QBrush(QColor("#ffaaaa"))
            standard = False
        if standard:
            QItemDelegate.paint(self, painter, option, index)
            return
        # custom render
        painter.save()
        painter.fillRect(option.rect, brush)
        data = self.model.data(index, Qt.DisplayRole)
        self.drawDisplay(painter, option, option.rect, data.toString())
        painter.restore()

class EditDeck(QDialog):

    def __init__(self, parent):
        QDialog.__init__(self, parent, Qt.Window)
        self.parent = parent
        self.config = parent.config
        self.dialog = ankiqt.forms.cardlist.Ui_EditDeck()
        self.dialog.setupUi(self)
        self.model = DeckModel(self.parent.deck)
        self.dialog.tableView.setSortingEnabled(False)
        self.dialog.tableView.setModel(self.model)
        self.dialog.tableView.setItemDelegate(StatusDelegate(self, self.model))
        self.dialog.tableView.selectionModel()
        self.dialog.tableView.setFont(QFont(
            self.config['editFontFamily'],
            self.config['editFontSize']))
        self.dialog.tableView.update()
        self.setupButtons()
        self.setupFilter()
        self.setupSort()
        self.setupHeaders()
        self.setupEditor()
        self.setupCardInfo()
        self.selectLastCard()
        self.dialog.filterEdit.setFocus()
        ui.dialogs.open("CardList", self)
        self.show()

    def selectLastCard(self):
        "Show the row corresponding to the current card."
        # this assumes all lines are being shown
        if self.parent.currentCard:
            row = self.model.cardIndex(self.parent.currentCard)
            idx = self.model.index(row, 0)
            self.dialog.tableView.selectionModel().setCurrentIndex(
                idx, QItemSelectionModel.Select | QItemSelectionModel.Rows)

    def setupFilter(self):
        self.filterTimer = None
        self.currentTag = None
        self.connect(self.dialog.filterEdit,
                     SIGNAL("textChanged(QString)"),
                     self.filterTextChanged)
        self.updateFilterLabel()
        self.setTabOrder(self.dialog.filterEdit, self.dialog.tableView)
        self.connect(self.dialog.tagList, SIGNAL("activated(int)"),
                     self.tagChanged)

    def setupSort(self):
        self.sortIndex = 4
        self.sortRev = False
        self.drawSort()
        self.connect(self.dialog.sortBox, SIGNAL("activated(int)"),
                     self.sortChanged)
        self.connect(self.dialog.reverse, SIGNAL("stateChanged(int)"),
                     self.sortRevChanged)
        self.sortChanged(self.sortIndex)

    def drawTags(self):
        tags = self.parent.deck.allTags()
        self.alltags = tags['all']
        self.alltags.sort()
        self.dialog.tagList.clear()
        self.dialog.tagList.addItems(QStringList(
            ['All tags'] + self.alltags))
        if self.currentTag:
            try:
                idx = self.alltags.index(self.currentTag) + 1
            except ValueError:
                idx = 0
            self.dialog.tagList.setCurrentIndex(idx)

    def drawSort(self):
        self.sortList = [
            _("Creation date"),
            _("Modified date"),
            _("Due date"),
            _("Correct ratio"),
            ]
        self.sortFields = sorted(self.parent.deck.facts.allFields())
        self.sortList.extend([_("Field '%s'") % f for f in self.sortFields])
        self.dialog.sortBox.clear()
        self.dialog.sortBox.addItems(QStringList(self.sortList))
        self.dialog.sortBox.setCurrentIndex(self.sortIndex)

    def sortChanged(self, idx):
        if idx == 0:
            self.sortKey = ("attr", "created")
        elif idx == 1:
            self.sortKey = ("attr", "modified")
        elif idx == 2:
            self.sortKey = ("attr", "nextTime")
        elif idx == 3:
            self.sortKey = ("special", "ease")
        else:
            self.sortKey = ("field", self.sortFields[idx-4])
        self.sortIndex = idx
        self.doSort()

    def doSort(self):
        self.model.sort(*(self.sortKey + (self.sortRev,)))

    def sortRevChanged(self, idx):
        if idx == Qt.Checked:
            self.sortRev = True
        else:
            self.sortRev = False
        self.doSort()

    def tagChanged(self, idx):
        if idx == 0:
            self.currentTag = None
        else:
            self.currentTag = self.alltags[idx-1]
        self.updateSearch()

    def updateFilterLabel(self):
        self.setWindowTitle(_("Anki - Edit Deck (%(cur)d "
                              "of %(tot)d cards shown)") %
                            {"cur": len(self.model.deck),
                             "tot": len(self.parent.deck)})

    def filterTextChanged(self):
        interval = 200
        if self.filterTimer:
            self.filterTimer.setInterval(interval)
        else:
            self.filterTimer = QTimer(self)
            self.filterTimer.setSingleShot(True)
            self.filterTimer.start(interval)
            self.connect(self.filterTimer, SIGNAL("timeout()"), self.updateSearch)

    def updateSearch(self):
        self.model.showMatching(unicode(self.dialog.filterEdit.text()),
                                self.currentTag)
        self.updateFilterLabel()
        self.filterTimer = None
        if self.model.deck:
            self.dialog.cardInfoGroup.show()
            self.dialog.fieldsArea.show()
            self.dialog.tableView.selectionModel().setCurrentIndex(
                self.model.index(0, 0),
                QItemSelectionModel.Select | QItemSelectionModel.Rows)
        else:
            self.dialog.cardInfoGroup.hide()
            self.dialog.fieldsArea.hide()

    def setupHeaders(self):
        self.dialog.tableView.verticalHeader().hide()
        self.dialog.tableView.horizontalHeader().hide()
        for i in range(2):
            self.dialog.tableView.horizontalHeader().setResizeMode(i, QHeaderView.Stretch)
        self.dialog.tableView.horizontalHeader().setResizeMode(2, QHeaderView.Fixed)
        self.dialog.tableView.setColumnWidth(2, 100)

    def setupButtons(self):
        self.connect(self.dialog.action_Delete_card,
                     SIGNAL("triggered()"),
                     self.toggleDelete)
        self.connect(self.dialog.actionAdd_fact_tag,
                     SIGNAL("triggered()"),
                     self.addFactTag)
        self.connect(self.dialog.actionAdd_card_tag,
                     SIGNAL("triggered()"),
                     self.addCardTag)
        self.connect(self.dialog.actionDelete_fact_tag,
                     SIGNAL("triggered()"),
                     self.deleteFactTag)
        self.connect(self.dialog.actionDelete_card_tag,
                     SIGNAL("triggered()"),
                     self.deleteCardTag)
        self.connect(self.dialog.actionMakeDue,
                     SIGNAL("triggered()"),
                     self.makeDue)
        self.connect(self.dialog.action_Reset_Progress,
                     SIGNAL("triggered()"),
                     self.resetProgress)
        self.connect(self.dialog.actionAdd_Missing_Cards,
                     SIGNAL("triggered()"),
                     self.addMissingCards)
        self.connect(self.dialog.actionDelete_Fact,
                     SIGNAL("triggered()"),
                     self.deleteFact)

    def contextMenuEvent(self, evt):
        menu = QMenu(self)
        factmenu = QMenu(_("For each fact..."), menu)
        menu.addMenu(factmenu)
        factmenu.addAction(self.dialog.actionAdd_fact_tag)
        factmenu.addAction(self.dialog.actionDelete_fact_tag)
        factmenu.addSeparator()
        factmenu.addAction(self.dialog.actionAdd_Missing_Cards)
        factmenu.addSeparator()
        factmenu.addAction(self.dialog.actionDelete_Fact)
        menu.addSeparator()
        menu.addAction(self.dialog.actionAdd_card_tag)
        menu.addAction(self.dialog.actionDelete_card_tag)
        menu.addSeparator()
        menu.addAction(self.dialog.actionMakeDue)
        menu.addAction(self.dialog.action_Reset_Progress)
        menu.addSeparator()
        menu.addAction(self.dialog.action_Delete_card)
        menu.exec_(evt.globalPos())

    def toggleDelete(self):
        cards = self.selectedCards()
        self.dialog.tableView.selectionModel().blockSignals(True)
        self.dialog.tableView.selectionModel().clear()
        self.dialog.tableView.selectionModel().blockSignals(False)
        for card in cards:
            self.model.deleted.symmetric_difference_update([card])
            self.model.emit(SIGNAL("layoutChanged()"))

    def selectedCards(self):
        return [self.model.deck[idx.row()] for idx in
                self.dialog.tableView.selectionModel().selectedRows()]

    def selectedFacts(self):
        facts = {}
        for c in self.selectedCards():
            if c.fact.id not in facts:
                facts[c.fact.id] = c.fact
        return facts.values()

    def setupEditor(self):
        self.editor = ui.facteditor.FactEditor(self,
                                               self.dialog.fieldsArea,
                                               self.parent.deck)
        self.editor.onChange = self.textChanged
        self.editor.onFactValid = self.onFactValid
        self.editor.onFactInvalid = self.onFactInvalid
        self.connect(self.dialog.tableView.selectionModel(),
                     SIGNAL("currentRowChanged(QModelIndex, QModelIndex)"),
                     self.rowChanged)

    def onFactValid(self, fact):
        self.dialog.tableView.setEnabled(True)
        self.dialog.searchGroup.setEnabled(True)
        self.dialog.sortGroup.setEnabled(True)

    def onFactInvalid(self, fact):
        self.dialog.tableView.setEnabled(False)
        self.dialog.searchGroup.setEnabled(False)
        self.dialog.sortGroup.setEnabled(False)

    def rowChanged(self, current, previous):
        self.currentRow = current
        self.currentCard = self.model.getCard(current)
        self.drawTags()
        fact = self.currentCard.fact
        self.editor.setFact(fact, True)
        self.showCardInfo(self.currentCard)

    def setupCardInfo(self):
        self.currentCard = None
        self.cardInfoGrid = QGridLayout(self.dialog.cardInfoGroup)
        self.cardInfoGrid.setMargin(6)
        # card
        l = QLabel(_("<b>Tags</b>"))
        self.cardInfoGrid.addWidget(l, 0, 0)
        self.cardStaticTags = QLabel()
        self.cardStaticTags.setWordWrap(True)
        self.cardInfoGrid.addWidget(self.cardStaticTags, 1, 0)
        l = QLabel(_("<b>Card-specific tags</b>"))
        self.cardInfoGrid.addWidget(l, 2, 0)
        self.cardTags = QLineEdit()
        self.connect(self.cardTags, SIGNAL("textChanged(QString)"), self.saveCardInfo)
        self.cardInfoGrid.addWidget(self.cardTags, 3, 0)
        l = QLabel(_("<b>Statistics</b>"))
        self.cardInfoGrid.addWidget(l, 4, 0)
        self.cardStats = QLabel()
        self.cardInfoGrid.addWidget(self.cardStats, 5, 0)
        item = QSpacerItem(20, 20, QSizePolicy.Expanding,
                           QSizePolicy.Expanding)
        self.cardInfoGrid.addItem(item, 6, 0)

    def updateStaticTags(self):
        card = self.currentCard
        self.cardStaticTags.setText(
            ", ".join(
            card.fact.model.tags +
            [card.model.name] +
            card.fact.tags))

    def showCardInfo(self, card):
        self.cardTags.setText(", ".join(card.tags))
        self.updateStaticTags()
        # stats
        next = time.time() - card.nextTime
        if next > 0:
            next = "%s ago" % anki.utils.fmtTimeSpan(next)
        else:
            next = "in %s" % anki.utils.fmtTimeSpan(abs(next))
        stats = self.parent.deck.sched.summarizeStats(card.stats)
        self.cardStats.setText(
            _("Created: %(c)s ago<br>"
              "Next due: %(n)s<br>"
              "Interval: %(i)0.0f days<br>"
              "Average: %(a)s<br>"
              "Total: %(t)s<br>"
              "Reviews: %(cor)d/%(tot)d<br>"
              "Succesive: %(suc)d")% {
            "c": fmtTimeSpan(time.time() - card.created),
            "n": next,
            "i": card.interval,
            "a": fmtTimeSpan(card.stats['averageTime']),
            "t": fmtTimeSpan(card.stats['totalTime']),
            "cor": stats['YesTotal'],
            "tot": stats['Total'],
            "suc": card.stats['successivelyCorrect'],
            })

    def saveCardInfo(self, text):
        if self.currentCard:
            tags = parseTags(unicode(text))
            if self.currentCard.tags != tags:
                self.currentCard.tags = tags
        self.drawTags()

    def textChanged(self, field):
        self.updateStaticTags()
        self.model.emit(SIGNAL("layoutChanged()"))
        self.parent.deck.setModified()

    def accept(self):
        if getattr(self, "alreadyClosed", None):
            return
        self.alreadyClosed = True
        deck = self.parent.deck
        for card in self.model.deleted:
            deck.remove(card)
        self.parent.setStatus(
            _("%(del)d deleted.") %
            {"del": len(self.model.deleted)})
        self.parent.deck.updatePriorities()
        self.parent.moveToState("deckOpened")
        ui.dialogs.close("CardList")
        self.close()

    def reject(self):
        self.accept()

    def addFactTag(self):
        tag = ui.utils.getText(_("Enter tag to add to each fact:"), self)
        if tag:
            for card in self.selectedCards():
                addTags(tag, card.fact.tags)
                card.fact.setModified()
        self.updateAfterCardChange()

    def addCardTag(self):
        tag = ui.utils.getText(_("Enter tag to add to each card:"), self)
        if tag:
            for card in self.selectedCards():
                addTags(tag, card.tags)
                card.setModified()
        self.updateAfterCardChange()

    def deleteFactTag(self):
        tag = ui.utils.getText(_("Enter tag to delete from each fact:"), self)
        if tag:
            for card in self.selectedCards():
                deleteTags(tag, card.fact.tags)
                card.fact.setModified()
        self.updateAfterCardChange()

    def deleteCardTag(self):
        tag = ui.utils.getText(_("Enter tag to delete from each card:"), self)
        if tag:
            for card in self.selectedCards():
                deleteTags(tag, card.tags)
                card.setModified()
        self.updateAfterCardChange()

    def updateAfterCardChange(self):
        self.rowChanged(self.currentRow, None)

    def resetProgress(self):
        if ui.utils.askUser(_(
            "Are you sure you wish to lose all progress "
            "on the selected cards?"), self):
            for card in self.selectedCards():
                self.parent.deck.sched.resetCard(card)
            self.model.reset()

    def makeDue(self):
        if ui.utils.askUser(_(
            "Are you sure you wish to reset the last iteration "
            "on the selected cards?"), self):
            for card in self.selectedCards():
                if not self.parent.deck.sched.cardIsPending(card):
                    self.parent.deck.sched.makeDue(card)
            self.model.reset()

    def addMissingCards(self):
        def haveModel(fact, model):
            for card in fact.cards:
                if card.model == model:
                    return True
            return False
        for fact in self.selectedFacts():
            for model in fact.model.cards:
                if not haveModel(fact, model):
                    # FIXME: should probably put it in the right place
                    card = model.makeCard(fact)
                    self.parent.deck.addCard(card,
                                             len(self.parent.deck))
        # length of deck changed. need to update
        self.updateSearch()

    def deleteFact(self):
        # FIXME: make this consistent with current deleting behaviour
        if ui.utils.askUser(_(
            "Even non-selected cards will be deleted if they share the\n"
            "same fact as a selected card. Are you sure?"), self):
            for fact in self.selectedFacts():
                for card in fact.cards[:]:
                    self.parent.deck.remove(card)
            self.updateSearch()
