# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

from PyQt4.QtGui import *
from PyQt4.QtCore import *
import sys, re
import ankiqt.forms
import anki
from ankiqt import ui
from anki.utils import parseTags

class DeckProperties(QDialog):

    def __init__(self, parent):
        QDialog.__init__(self, parent, Qt.Window)
        self.parent = parent
        self.dialog = ankiqt.forms.deckproperties.Ui_DeckProperties()
        self.dialog.setupUi(self)
        self.modified = False
        self.models = self.parent.deck.facts.models
        self.readData()
        self.connect(self.dialog.modelsAdd, SIGNAL("clicked()"), self.onAdd)
        self.connect(self.dialog.modelsEdit, SIGNAL("clicked()"), self.onEdit)
        self.connect(self.dialog.modelsDelete, SIGNAL("clicked()"), self.onDelete)
        self.show()

    def readData(self):
        # description
        self.dialog.deckDescription.setText(self.parent.deck.description)
        # syncing
        sn = self.parent.deck.syncName
        if sn:
            self.dialog.doSync.setCheckState(Qt.Checked)
            self.dialog.syncName.setText(sn)
        else:
            self.dialog.doSync.setCheckState(Qt.Unchecked)
            self.dialog.syncName.setText(self.parent.deck.name())
        # priorities
        self.dialog.highPriority.setText(", ".join(
            self.parent.deck.sched.highPriority))
        self.dialog.medPriority.setText(", ".join(
            self.parent.deck.sched.medPriority))
        self.dialog.lowPriority.setText(", ".join(
            self.parent.deck.sched.lowPriority))
        self.dialog.postponing.setText(", ".join(
            self.parent.deck.sched.suspendedTags))
        # scheduling
        minmax = ("Min", "Max")
        for type in ("hard", "mid", "easy"):
            v = getattr(self.parent.deck.sched, type + "Interval")
            for n in range(len(minmax)):
                getattr(self.dialog, type + minmax[n]).setText("%0.1f" % v[n])
        self.dialog.delay0.setText(unicode(self.parent.deck.sched.delay0/60.0))
        self.dialog.delay1.setText(unicode(self.parent.deck.sched.delay1/60.0))
        self.dialog.delay2.setText(unicode(self.parent.deck.sched.delay2/60.0))
        self.dialog.collapse.setText(unicode(self.parent.deck.sched.collapse/60.0/60.0))
        self.dialog.intervalCap.setText(unicode(
            self.parent.deck.sched.maxScheduleTime/365.0))
        # models
        self.updateModelsList()

    def updateModelsList(self):
        self.dialog.modelsList.clear()
        for model in self.models:
            name = _("%(name)s [%(facts)d facts, %(cards)d cards]") % {
                'name': model.name,
                'facts': self.parent.deck.facts.modelUseCount(model),
                'cards': self.parent.deck.modelUseCount(model),
                }
            item = QListWidgetItem(name)
            self.dialog.modelsList.addItem(item)
            if model.name == self.parent.deck.facts.models.current.name:
                self.dialog.modelsList.setCurrentItem(item)

    def onAdd(self):
        m = ui.modelchooser.AddModel(self, self.parent).getModel()
        if m:
            self.parent.deck.facts.models.setModel(m)
            self.updateModelsList()
            self.modified = True

    def onEdit(self):
        model = self.selectedModel()
        if not model:
            return
        ui.modelproperties.ModelProperties(self, model, self.parent)
        self.updateModelsList()
        self.modified = True

    def onDelete(self):
        model = self.selectedModel()
        if not model:
            return
        if len(self.parent.deck.facts.models) < 2:
            ui.utils.showWarning(_("You can't delete the last model.\n"
                                   "Please add another model first."),
                                 parent=self)
            return
        count = self.parent.deck.modelUseCount(model)
        if count:
            if not ui.utils.askUser(
                _("This model is used by %d cards.\n"
                  "Are you sure you want to delete it?\n"
                  "If you delete it, these cards will be lost.")
                % count, parent=self):
                return
        self.parent.deck.removeModel(model)
        self.updateModelsList()
        self.modified = True

    def selectedModel(self):
        row = self.dialog.modelsList.currentRow()
        if row == -1:
            return None
        return self.models[self.dialog.modelsList.currentRow()]

    def updateField(self, obj, field, value):
        if getattr(obj, field) != value:
            setattr(obj, field, value)
            self.modified = True

    def reject(self):
        # description
        self.updateField(self.parent.deck, 'description',
                         unicode(self.dialog.deckDescription.toPlainText()))
        # syncing
        if self.dialog.doSync.checkState() == Qt.Checked:
            self.updateField(self.parent.deck, 'syncName',
                             unicode(self.dialog.syncName.text()))
        else:
            self.updateField(self.parent.deck, 'syncName', None)
        # scheduling
        minmax = ("Min", "Max")
        for type in ("hard", "mid", "easy"):
            for n in range(len(minmax)):
                v = getattr(self.dialog, type + minmax[n]).text()
                try:
                    v = float(v)
                except ValueError:
                    # FIXME: tell the user instead of silently ignoring
                    continue
                l = getattr(self.parent.deck.sched, type + "Interval")
                l = list(l)
                if l[n] != v:
                    self.modified = True
                l[n] = v
                setattr(self.parent.deck.sched, type + "Interval", l)
        try:
            v = float(self.dialog.intervalCap.text()) * 365.0
            self.updateField(self.parent.deck.sched, 'maxScheduleTime', v)
            v = float(self.dialog.delay0.text()) * 60.0
            self.updateField(self.parent.deck.sched, 'delay0', v)
            v = float(self.dialog.delay1.text()) * 60.0
            self.updateField(self.parent.deck.sched, 'delay1', v)
            v = float(self.dialog.delay2.text()) * 60.0
            self.updateField(self.parent.deck.sched, 'delay2', v)
            v = float(self.dialog.collapse.text()) * 60.0 * 60.0
            self.updateField(self.parent.deck.sched, 'collapse', v)
        except ValueError:
            pass
        self.updateField(self.parent.deck.sched,
                         "highPriority",
                         parseTags(unicode(self.dialog.highPriority.text())))
        self.updateField(self.parent.deck.sched,
                         "medPriority",
                         parseTags(unicode(self.dialog.medPriority.text())))
        self.updateField(self.parent.deck.sched,
                         "lowPriority",
                         parseTags(unicode(self.dialog.lowPriority.text())))
        self.updateField(self.parent.deck.sched,
                         "suspendedTags",
                         parseTags(unicode(self.dialog.postponing.text())))
        # mark deck dirty and close
        if self.modified:
            self.parent.deck.updatePriorities()
            self.parent.deck.setModified()
            self.parent.moveToState("deckOpened")
        self.close()
