# Copyright: Damien Elmes <anki@ichi2.net>
# -*- coding: utf-8 -*-
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

from PyQt4.QtGui import *
from PyQt4.QtCore import *

import os, sys, re, types, gettext, stat, traceback, shutil, time

from PyQt4.QtCore import *
from PyQt4.QtGui import *
import anki, anki.sync, anki.stats, anki.lang
import anki.utils, anki.models, anki.stdmodels
import anki.fonts
from anki.errors import *
from anki.sound import hasSound, playFromText
import ankiqt
ui = ankiqt.ui
config = ankiqt.config

class AnkiQt(QMainWindow):
    def __init__(self, app, config, args):
        QMainWindow.__init__(self)
        ankiqt.mw = self
        self.app = app
        self.config = config
        self.deck = None
        self.views = []
        self.currentCard = None
        self.macHacks()
        self.setLang()
        self.setupFonts()
        self.setupBackupDir()
        self.setupHooks()
        self.loadUserCustomisations()
        self.mainWin = ankiqt.forms.main.Ui_MainWindow()
        self.mainWin.setupUi(self)
        self.alterShortcuts()
        self.help = ui.help.HelpArea(self.mainWin.helpFrame, self.config, self)
        self.connectMenuActions()
        if sys.platform.startswith("darwin"):
            # window creeps on osx, just resize
            self.resize(*self.config['mainWindowGeometry'][2:4])
        else:
            self.setGeometry(QRect(*self.config['mainWindowGeometry']))
        self.bodyView = ui.view.View(self, self.mainWin.mainText,
                                     self.mainWin.mainTextFrame)
        self.addView(self.bodyView)
        self.statusView = ui.status.StatusView(self)
        self.addView(self.statusView)
        self.checkLock()
        self.setupButtons()
        self.deckStorage = anki.Controller.DeckStorage()
        if not self.config['showToolbar']:
            self.removeToolBar(self.mainWin.toolBar)
            self.mainWin.toolBar.hide()
        self.show()
        # run after-init hook. we need to catch exceptions here or it'll
        # prevent anki from starting
        try:
            self.runHook('init')
        except:
            print "Error running initHook in ~/.anki.py"
            print traceback.print_exc()
        self.maybeLoadLastDeck(args)
        self.setupAutoUpdate()

    # State machine
    ##########################################################################

    def addView(self, view):
        self.views.append(view)

    def updateViews(self, status):
        if self.deck is None and status != "noDeck":
            raise "updateViews() called with no deck. status=%s" % status
        for view in self.views:
            view.setState(status)

    def moveToState(self, state):
        # check if we have an open deck and respond appropriately
        if state == "initial":
            self.currentCard = None
            if self.deck == None:
                return self.moveToState("noDeck")
            else:
                return self.moveToState("deckOpened")
        # save the new & last state
        self.lastState = getattr(self, "state", None)
        self.state = state
        self.updateTitleBar()
        if state == "deckOpened":
            if self.deck.mediaDir():
                os.chdir(self.deck.mediaDir())
            self.deck.sched.resetPending()
            self.mainWin.menu_Lookup.setEnabled(True)
            self.enableDeckMenuItems()
            self.updateRecentFilesMenu()
            self.updateViews(state)
            # save regenerating the pending list if we can
            if (self.currentCard and
                self.deck.sched.cardIsPending(self.currentCard) and
                self.lastState in ("showQuestion", "showAnswer") and
                self.state in ("showQuestion", "showAnswer")):
                self.state = self.lastState
                return self.moveToState(self.lastState)
            self.currentCard = None
            return self.moveToState("getQuestion")
        elif state == "noDeck":
            self.help.hide()
            self.disableDeckMenuItems()
            self.resetButtons()
        elif state == "getQuestion":
            self.lastCard = self.currentCard
            if len(self.deck) == 0:
                return self.moveToState("deckEmpty")
            else:
                card = self.deck.sched.getCard()
                if card:
                    self.currentCard = card
                    self.enableCardMenuItems()
                    return self.moveToState("showQuestion")
                else:
                    return self.moveToState("deckFinished")
        elif state == "deckEmpty":
            self.resetButtons()
            self.disableCardMenuItems()
            self.mainWin.menu_Lookup.setEnabled(False)
        elif state == "deckFinished":
            self.resetButtons()
            self.mainWin.menu_Lookup.setEnabled(False)
            self.disableCardMenuItems()
            self.startRefreshTimer()
            self.bodyView.setState(state)
        elif state == "showQuestion":
            self.resetButtons()
            self.showAnswerButton()
            self.updateMarkAction()
            self.runHook('showQuestion')
        elif state == "showAnswer":
            self.resetButtons()
            self.showEaseButtons()
            self.enableCardMenuItems()
        self.updateViews(state)

    def keyPressEvent(self, evt):
        "Show answer on RET or register answer."
        if self.state == "showQuestion":
            if evt.key() in (Qt.Key_Enter,
                             Qt.Key_Return):
                evt.accept()
                return self.moveToState("showAnswer")
        elif self.state == "showAnswer":
            key = unicode(evt.text())
            if key and key >= "0" and key <= "4":
                # user entered a quality setting
                num=int(key)
                evt.accept()
                return self.cardAnswered(num)
        evt.ignore()

    def cardAnswered(self, quality):
        "Reschedule current card and move back to getQuestion state."
        n = self.deck.sched.answerCard(self.currentCard, quality)
        if n:
            self.lastScheduledTime = anki.utils.fmtTimeSpan(n*86400)
        else:
            self.lastScheduledTime = 0
        if self.config['saveAfterAnswer']:
            num = self.config['saveAfterAnswerNum']
            stats = self.deck.sched.getStats()
            if stats['sTotal'] % num == 0:
                self.save(backup=False)
        self.moveToState("getQuestion")

    def startRefreshTimer(self):
        "Update the screen once a minute until next card is displayed."
        if getattr(self, 'refreshTimer', None):
            return
        self.refreshTimer = QTimer(self)
        self.refreshTimer.start(60000)
        self.connect(self.refreshTimer, SIGNAL("timeout()"), self.refreshStatus)
        # start another time to refresh exactly after we've finished
        next = self.deck.sched.earliestTime()
        if next:
            delay = next - time.time()
            delay = (delay+1)*1000
            if delay > 86400:
                return
            t = QTimer(self)
            t.setSingleShot(True)
            self.connect(t, SIGNAL("timeout()"), self.refreshStatus)
            t.start((delay+1)*1000)

    def refreshStatus(self):
        "If triggered when the deck is finished, reset state."
        if self.state == "deckFinished":
            self.moveToState("deckOpened")
        if self.state != "deckFinished":
            self.refreshTimer.stop()
            self.refreshTimer = None

    # Buttons
    ##########################################################################

    def setupButtons(self):
        self.outerButtonBox = QHBoxLayout(self.mainWin.buttonWidget)
        self.outerButtonBox.setMargin(3)
        self.outerButtonBox.setSpacing(0)
        self.innerButtonWidget = None

    def resetButtons(self):
        # this round-about process is trying to work around a bug in qt
        if self.lastState == self.state:
            return
        if self.innerButtonWidget:
            self.outerButtonBox.removeWidget(self.innerButtonWidget)
            self.innerButtonWidget.setParent(None)
        self.innerButtonWidget = QWidget()
        self.outerButtonBox.addWidget(self.innerButtonWidget)
        self.buttonBox = QVBoxLayout(self.innerButtonWidget)
        self.buttonBox.setSpacing(3)
        self.buttonBox.setMargin(3)
        if self.config['easeButtonHeight'] == "tall":
            self.easeButtonHeight = 50
        else:
            if sys.platform.startswith("darwin"):
                self.easeButtonHeight = 35
            else:
                self.easeButtonHeight = 25

    def showAnswerButton(self):
        if self.lastState == self.state:
            return
        button = QPushButton(_("Show answer"))
        button.setFixedHeight(self.easeButtonHeight)
        self.buttonBox.addWidget(button)
        button.setFocus()
        button.setDefault(True)
        self.connect(button, SIGNAL("clicked()"),
                     lambda: self.moveToState("showAnswer"))

    def getSpacer(self, hpolicy=QSizePolicy.Preferred):
        return QSpacerItem(20, 20,
                           hpolicy,
                           QSizePolicy.Preferred)

    def showEaseButtons(self):
        # if the state hasn't changed, do nothing
        if self.lastState == self.state:
            return
        # gather next intervals
        nextInts = {}
        for i in range(5):
            s=self.deck.sched.nextIntervalStr(self.currentCard, i)
            nextInts["ease%d" % i] = s
        # young card message
        if self.deck.sched.cardIsInInitialState(self.currentCard):
            text = (
                (_("Completely forgot."), ""),
                (_("Made a mistake."), ""),
                (_("Want to practice again tomorrow."),
                 _("Next in <b>%(ease2)s</b>")),
                (_("Want to practice again in a few days."),
                 _("Next in <b>%(ease3)s</b>")),
                (_("Remembered easily - ready to level up."),
                 _("Next in <b>%(ease4)s</b>")))
        else:
            # old card message
            text = (
                (_("Completely forgot. Reset progress."), ""),
                (_("Made a mistake. Back two steps."), ""),
                (_("Barely remembered. Back one step."),
                 _("Next in <b>%(ease2)s</b>")),
                (_("Remembered after thinking for a while."),
                 _("Next in <b>%(ease3)s</b>")),
                (_("Remembered easily."),
                 _("Next in <b>%(ease4)s</b>")))
        # button grid
        grid = QGridLayout()
        grid.setSpacing(3)
        if self.config['easeButtonStyle'] == 'standard':
            button3 = self.showStandardEaseButtons(grid, nextInts, text)
        else:
            button3 = self.showCompactEaseButtons(grid, nextInts)
        self.buttonBox.addLayout(grid)
        button3.setFocus()

    def showStandardEaseButtons(self, grid, nextInts, text):
        # show 'how well?' message
        hbox = QHBoxLayout()
        hbox.addItem(self.getSpacer(QSizePolicy.Expanding))
        label = QLabel(self.withInterfaceFont(
            _("<b>How well did you remember?</b>")))
        hbox.addWidget(label)
        hbox.addItem(self.getSpacer(QSizePolicy.Expanding))
        self.buttonBox.addLayout(hbox)
        # populate buttons
        button3 = None
        for i in range(5):
            button = QPushButton(str(i))
            button.setFixedWidth(100)
            button.setFixedHeight(self.easeButtonHeight)
            if i == 3:
                button3 = button
            grid.addItem(self.getSpacer(QSizePolicy.Expanding), i, 0)
            grid.addWidget(button, i, 1)
            grid.addItem(self.getSpacer(), i, 2)
            grid.addWidget(QLabel(self.withInterfaceFont(text[i][0])), i, 3)
            grid.addItem(self.getSpacer(), i, 4)
            grid.addWidget(QLabel(self.withInterfaceFont(
                text[i][1] % nextInts)), i, 5)
            grid.addItem(self.getSpacer(QSizePolicy.Expanding), i, 6)
            self.connect(button, SIGNAL("clicked()"),
                         lambda i=i: self.cardAnswered(i))
        return button3

    def showCompactEaseButtons(self, grid, nextInts):
        text = (
            (_("<b>%(ease0)s</b>")),
            (_("<b>%(ease1)s</b>")),
            (_("<b>%(ease2)s</b>")),
            (_("<b>%(ease3)s</b>")),
            (_("<b>%(ease4)s</b>")))
        button3 = None
        for i in range(5):
            button = QPushButton(str(i))
            button.setFixedHeight(self.easeButtonHeight)
            #button.setFixedWidth(70)
            if i == 3:
                button3 = button
            grid.addWidget(button, 0, (i*2)+1)
            label = QLabel(self.withInterfaceFont(text[i] % nextInts))
            label.setAlignment(Qt.AlignHCenter)
            grid.addWidget(label, 1, (i*2)+1)
            self.connect(button, SIGNAL("clicked()"),
                lambda i=i: self.cardAnswered(i))
        return button3

    def withInterfaceFont(self, text):
        family = self.config["interfaceFontFamily"]
        size = self.config["interfaceFontSize"]
        colour = self.config["interfaceColour"]
        css = ('.interface {font-family: "%s"; font-size: %spx; color: %s}\n' %
               (family, size, colour))
        css = "<style>\n" + css + "</style>\n"
        text = css + '<span class="interface">' + text + "</span>"
        return text

    # Hooks
    ##########################################################################

    def setupHooks(self):
        self.hooks = {}

    def addHook(self, hookName, func):
        if not self.hooks.get(hookName, None):
            self.hooks[hookName] = []
        if func not in self.hooks[hookName]:
            self.hooks[hookName].append(func)

    def removeHook(self, hookName, func):
        hook = self.hooks.get(hookName, [])
        if func in hook:
            hook.remove(func)

    def runHook(self, hookName, *args):
        hook = self.hooks.get(hookName, None)
        if hook:
            for func in hook:
                func(*args)

    # Deck loading & saving: backend
    ##########################################################################

    def setupBackupDir(self):
        anki.Controller.backupDir = os.path.join(
            self.config.configPath, "backups")

    def loadDeck(self, deckPath):
        "Load a deck and update the user interface. Maybe sync."
        try:
            self.deck = self.deckStorage.load(deckPath)
        except (IOError, ImportError):
            return False
        except DeckWrongFormatError:
            self.importOldDeck(deckPath)
            if not self.deck:
                return
        # backup on load, since we may automatically save
        self.deckStorage.backup(deckPath)
        self.moveToState("deckOpened")
        if self.config['syncOnLoad']:
            self.syncDeck(False)
        return True

    def loadRecent(self, n):
        if not self.saveAndClose(): return
        path = self.config['recentDeckPaths'][n]
        if self.loadDeck(path):
            self.updateRecentFiles(path)

    def importOldDeck(self, deckPath):
        ui.utils.showInfo(_("""
<h1>Upgrade deck</h1>Anki
0.3 uses a new deck format. To use your current deck with Anki, it needs to be
upgraded.
<p>
Please read the <a
href="http://ichi2.net/anki/help/upgrade0.3.html">upgrade notes</a> before
proceeding.
<p>
The upgrade process may take a while.""".strip()))
        if self.config.get('targetLang', "") == "en":
            model = anki.stdmodels.byName("English")
        else:
            model = anki.stdmodels.byName("Japanese")
        self.deck = None
        try:
            self.deck = anki.importing.upgradeV2(deckPath, model)
        except DeckWrongFormatError, e:
            ui.utils.showWarning(_(
                "An error occurred while upgrading:\n%s") % `e.data`)
            return
        self.save()

    def maybeLoadLastDeck(self, args):
        "Open the last deck if possible."
        # try a command line argument if available
        try:
            if args:
                file = unicode(args[0], sys.getfilesystemencoding())
                if self.loadDeck(file):
                    self.updateRecentFiles(file)
                    return
            elif self.config['recentDeckPaths']:
                deck = self.config['recentDeckPaths'][-1]
                # necessary for upgrading from 0.2
                try:
                    deck = unicode(deck, sys.getfilesystemencoding())
                except TypeError:
                    pass
                if self.loadDeck(deck):
                    return
        except:
            sys.stderr.write("Error loading last deck.\n")
            traceback.print_exc()
            self.deck = None
            return self.moveToState("initial")
        return self.moveToState("initial")

    def getDefaultDir(self, save=False):
        "Try and get default dir from most recently opened file."
        defaultDir = ""
        if self.config['recentDeckPaths']:
            latest = self.config['recentDeckPaths'][-1]
            defaultDir = os.path.dirname(latest)
        else:
            if save:
                defaultDir = unicode(os.path.expanduser("~/"),
                                     sys.getfilesystemencoding())
            else:
                samples = self.getSamplesDir()
                if samples:
                    return samples
        return defaultDir

    def getSamplesDir(self):
        path = os.path.join(ankiqt.runningDir, "libanki")
        if not os.path.exists(path):
            path = os.path.join(
                os.path.join(ankiqt.runningDir, ".."), "libanki")
            if not os.path.exists(path):
                path = ankiqt.runningDir
        if sys.platform.startswith("win32"):
            path = os.path.split(
                os.path.split(ankiqt.runningDir)[0])[0]
        elif sys.platform.startswith("darwin"):
            path = ankiqt.runningDir + "/../../.."
        else:
            path = os.path.join(path, "anki")
        path = os.path.join(path, "samples")
        path = os.path.normpath(path)
        if os.path.exists(path):
            if sys.platform.startswith("darwin"):
                return self.openMacSamplesDir(path)
            return path
        return ""

    def openMacSamplesDir(self, path):
        # some versions of macosx don't allow the open dialog to point inside
        # a .App file, it seems - so we copy the files onto the desktop.
        newDir = os.path.expanduser("~/Documents/Anki 0.3 Sample Decks")
        import shutil
        if os.path.exists(newDir):
            files = os.listdir(path)
            for file in files:
                loc = os.path.join(path, file)
                if not os.path.exists(os.path.join(newDir, file)):
                    shutil.copy2(loc, newDir)
            return newDir
        shutil.copytree(path, newDir)
        return newDir

    def save(self, backup=True, path=None):
        if self.deckHasPath() or path:
            if not path and self.deckIsReadOnly():
                self.setStatus(
                    _("Read-only. Please use Deck -> Save As"))
                return False
            self.setStatus(_("Saving.."))
            try:
                self.deckStorage.save(self.deck, backup=backup, path=path)
            except (OSError, IOError), e:
                ui.utils.showWarning(_("Unable to save file!\n%s") %
                                     unicode(e))
                self.setStatus(_("Saving failed"))
                return False
            else:
                self.setStatus(_("Saving..done"))
                self.updateRecentFiles(self.deck.path)
                # we may need to reenable save
                self.enableDeckMenuItems()
                self.updateTitleBar()
                return True
        else:
            self.setStatus(_("Please save the current file."))
            return False

    def deckHasPath(self):
        "True if the deck has been associated with a filename."
        return self.deck != None and \
               getattr(self.deck, "path", None)

    def updateRecentFiles(self, path):
        "Add the current deck to the list of recent files."
        path = os.path.normpath(path)
        if path in self.config['recentDeckPaths']:
            self.config['recentDeckPaths'].remove(path)
        self.config['recentDeckPaths'].append(path)
        if len(self.config['recentDeckPaths']) > 4:
            self.config['recentDeckPaths'] = \
              self.config['recentDeckPaths'][-4:]
        self.config.save()
        self.updateRecentFilesMenu()

    def deckIsReadOnly(self):
        if self.deckHasPath():
            try:
                st = os.stat(self.deck.path)
            except (IOError, OSError):
                return False # FIXME: test directory
            if st[stat.ST_MODE] & stat.S_IWUSR:
                return False
            else:
                return True
        else:
            return True

    # New files, loading & saving
    ##########################################################################

    def saveAndClose(self, exit=False):
        "(Auto)save and close. Prompt if necessary. True if okay to proceed."
        canSync = True
        # autosaving
        if self.deck is not None and self.deck.modifiedSinceSave():
            needPrompt = True
            # autosave if we have a path
            if self.config['saveOnClose'] and self.deckHasPath():
                if self.save():
                    # if auto-saving succeeded, no need to prompt
                    needPrompt = False
            if needPrompt:
                # otherwise, ask
                res = ui.unsaved.ask(self)
                if res == ui.unsaved.save:
                    # if manual-saving failed, fail
                    if not self.onSave():
                        return False
                elif res == ui.unsaved.cancel:
                    # if aborted by user, fail
                    return False
                elif res == ui.unsaved.discard:
                    # if the user discarded the save, don't sync either
                    canSync = False
        # syncing
        if self.config['syncOnClose'] and canSync:
            self.syncDeck(False)
        # close
        self.deck = None
        if not exit:
            self.moveToState("noDeck")
        return True

    def onNew(self):
        if not self.saveAndClose(): return
        self.newFile()
        if self.deck is not None and self.deck.facts.models:
            ui.dialogs.get("AddCards", self)

    def newFile(self):
        "Open a new file, prompting the user for the default model."
        self.deck = anki.Controller.Deck()
        # automatically associate a file name if it's their first deck
        firstDeck = unicode(os.path.expanduser("~/mydeck.anki"),
                            sys.getfilesystemencoding())
        if not os.path.exists(firstDeck):
            self.deck.path = firstDeck
        # remove the default model, and prompt
        del self.deck.facts.models[0]
        m = ui.modelchooser.AddModel(self, online=True).getModel()
        if m:
            if m != "online":
                self.deck.facts.models.setModel(m)
                self.deck.lastLoaded = self.deck.modified
                self.moveToState("deckOpened")
                return
            self.deck.syncName = ""
            # ensure all changes come to us
            self.deck.__dict__['modified'] = 0
            self.syncDeck(onlyMerge=True)
            return
        self.deck = None
        self.moveToState("initial")

    def onOpen(self, samples=False):
        # FIXME: refactor this and the saving code to use the utils functions
        if not self.saveAndClose(): return
        if samples:
            defaultDir = self.getSamplesDir()
            key = _("Deck files (*.anki)")
        else:
            defaultDir = self.getDefaultDir()
            key = _("Deck files (*.fc *.anki)")
        file = QFileDialog.getOpenFileName(self, _("Open deck"),
                                           defaultDir, key)
        file = unicode(file)
        if not file:
            return False
        if not self.loadDeck(file):
            ui.utils.showWarning(_("Unable to load file."))
            self.deck = None
            return False
        else:
            self.updateRecentFiles(file)
            return True

    def onOpenSamples(self):
        self.onOpen(samples=True)

    def onSave(self):
        "Save a file, prompting for a name if necessary. True on success."
        if self.deck.modifiedSinceSave():
            if self.deckHasPath():
                # existing file
                if self.save():
                    return True
            # no name yet or save failed
            return self.onSaveAs()
        else:
            self.setStatus(_("Deck is not modified."))
            return True

    def onSaveAs(self, title=None):
        "Prompt for a file name, then save. Return true if saved."
        if not title:
            title = _("Save deck")
        if self.deckHasPath():
            dir = os.path.dirname(self.deck.path)
        else:
            dir = self.getDefaultDir(save=True)
        file = QFileDialog.getSaveFileName(self, title,
                                           dir,
                                           _("Deck files (*.anki)"),
                                           None,
                                           QFileDialog.DontConfirmOverwrite)
        file = unicode(file)
        if not file:
            return False
        if not file.lower().endswith(".anki"):
            file += ".anki"
        if os.path.exists(file):
            # check for existence after extension
            if not ui.utils.askUser(
                "This file exists. Are you sure you want to overwrite it?"):
                return
        self.save(path=file)
        self.updateTitleBar()
        return True

    # Opening and closing the app
    ##########################################################################

    def prepareForExit(self):
        "Save config and window geometry."
        g = self.geometry()
        self.help.hide()
        self.config['mainWindowGeometry'] = (g.left(),
                                             g.top(),
                                             self.width(),
                                             self.height())
        # save config and kill main window
        try:
            self.config.save()
        except (IOError, OSError), e:
            ui.utils.showWarning(_("Anki was unable to save your "
                                   "configuration file:\n%s" % e))
        self.removeLock()

    def closeEvent(self, event):
        "User hit the X button, etc."
        if not self.saveAndClose(exit=True):
            event.ignore()
        else:
            self.prepareForExit()
            event.accept()
            self.app.quit()

    # Tools - looking up words in the dictionary
    ##########################################################################

    def initLookup(self):
        if not getattr(self, "lookup", None):
            self.lookup = ui.lookup.Lookup(self)

    def onLookupExpression(self):
        self.initLookup()
        try:
            self.lookup.alc(self.currentCard.fact['Expression'])
        except KeyError:
            self.setStatus(_("No expression in current card."))

    def onLookupMeaning(self):
        self.initLookup()
        try:
            self.lookup.alc(self.currentCard.fact['Meaning'])
        except KeyError:
            self.setStatus(_("No meaning in current card."))

    def onLookupEdictSelection(self):
        self.initLookup()
        self.lookup.selection(self.lookup.edict)

    def onLookupEdictKanjiSelection(self):
        self.initLookup()
        self.lookup.selection(self.lookup.edictKanji)

    def onLookupAlcSelection(self):
        self.initLookup()
        self.lookup.selection(self.lookup.alc)

    # Tools - statistics
    ##########################################################################

    def onKanjiStats(self):
        rep = anki.stats.KanjiStats(self.deck).report()
        rep += _("<a href=py:miss>Missing Kanji</a><br>")
        rep += _("<a href=py:occur>Kanji report</a>")
        self.help.showText(rep, py={"miss": self.onMissingStats,
                                    "occur": self.onKanjiOccur})

    def onMissingStats(self):
        ks = anki.stats.KanjiStats(self.deck)
        ks.genKanjiSets()
        self.help.showText(ks.missingReport())

    def onDeckStats(self):
        txt = anki.stats.DeckStats(self.deck).report()
        self.help.showText(txt)

    def onCardStats(self):
        self.addHook("showQuestion", self.onCardStats)
        self.addHook("helpChanged", self.removeCardStatsHook)
        txt = ""
        if self.currentCard:
            txt += _("<h1>Current card</h1>")
            txt += anki.stats.CardStats(self.currentCard).report()
        if self.lastCard and self.lastCard != self.currentCard:
            txt += _("<h1>Last card</h1>")
            txt += anki.stats.CardStats(self.lastCard).report()
        if not txt:
            txt = _("No current card or last card.")
        self.help.showText(txt, key="cardStats")

    def removeCardStatsHook(self):
        "Remove the update hook if the help menu was changed."
        if self.help.currentKey != "cardStats":
            self.removeHook("showQuestion", self.onCardStats)

    def onShowGraph(self):
        self.setStatus(_("Loading graphs (may take time).."))
        self.app.processEvents()
        import anki.graphs
        if anki.graphs.graphsAvailable():
            try:
                ui.dialogs.get("Graphs", self, self.deck)
            except (ImportError, ValueError):
                if sys.platform.startswith("win32"):
                    ui.utils.showInfo(
                        _("To display graphs, Anki needs a .dll file which\n"
                          "you don't have. Please install:\n") +
                        "http://www.dll-files.com/dllindex/dll-files.shtml?msvcp71")
                else:
                    ui.utils.showInfo(_(
                        "Your version of Matplotlib is broken.\n"
                        "Please see http://repose.ath.cx/tracker/anki/issue102"))
        else:
            ui.utils.showInfo(_("Please install python-matplotlib to access graphs."))

    def onKanjiOccur(self):
        self.setStatus(_("Generating report (may take time).."))
        self.app.processEvents()
        import tempfile
        (fd, name) = tempfile.mkstemp(suffix=".html")
        f = os.fdopen(fd, 'w')
        ko = anki.stats.KanjiOccurStats(self.deck)
        ko.reportFile(f)
        f.close()
        if sys.platform == "win32":
            url = "file:///"
        else:
            url = "file://"
        url += os.path.abspath(name)
        QDesktopServices.openUrl(QUrl(url))

    # Marking, suspending and undoing
    ##########################################################################

    def onMark(self, toggled):
        if self.currentCard.hasTag("Marked"):
            self.currentCard.tags.remove("Marked")
        else:
            self.currentCard.tags.append("Marked")
        self.currentCard.setModified()

    def onSuspend(self):
        if not self.currentCard.hasTag("Suspended"):
            self.currentCard.fact.tags.append("Suspended")
            self.currentCard.updatePriority()
        self.currentCard.setModified()
        self.lastScheduledTime = None
        self.moveToState("getQuestion")

    def onUndoAnswer(self):
        self.deck.sched.makeDue(self.lastCard)
        self.moveToState("deckOpened")

    # Other menu operations
    ##########################################################################

    def onAddCard(self):
        ui.dialogs.get("AddCards", self)

    def onEditDeck(self):
        ui.dialogs.get("CardList", self)

    def onDeckProperties(self):
        ui.deckproperties.DeckProperties(self)

    def onModelProperties(self):
        if self.currentCard:
            model = self.currentCard.fact.model
        else:
            model = self.deck.facts.models.current
        ui.modelproperties.ModelProperties(self, model)

    def onDisplayProperties(self):
        ui.dialogs.get("DisplayProperties", self)

    def onPrefs(self):
        ui.preferences.Preferences(self, self.config)

    def onReportBug(self):
        QDesktopServices.openUrl(QUrl(ankiqt.appIssueTracker))

    def onForum(self):
        QDesktopServices.openUrl(QUrl(ankiqt.appForum))

    def onAbout(self):
        ui.about.show(self)

    # Importing & exporting
    ##########################################################################

    def onImport(self):
        if self.deck is None:
            self.newFile()
        if self.deck is not None:
            ui.importing.ImportDialog(self)

    def onExport(self):
        ui.exporting.ExportDialog(self)

    # Language handling
    ##########################################################################

    def setLang(self):
        "Set the user interface language."
        languageDir=os.path.join(ankiqt.modDir, "locale")
        self.languageTrans = gettext.translation('ankiqt', languageDir,
                                            languages=[self.config["interfaceLang"]],
                                            fallback=True)
        self.installTranslation()
        if getattr(self, 'mainWin', None):
            self.mainWin.retranslateUi(self)
            self.alterShortcuts()
        anki.lang.setLang(self.config["interfaceLang"])
        self.updateTitleBar()

    def getTranslation(self, text):
        return self.languageTrans.ugettext(text)

    def installTranslation(self):
        import __builtin__
        __builtin__.__dict__['_'] = self.getTranslation

    # Syncing
    ##########################################################################

    def syncDeck(self, interactive=True, create=False, onlyMerge=False):
        "Synchronise a deck with the server."
        # vet input
        u=self.config['syncUsername']
        p=self.config['syncPassword']
        if not u or not p:
            msg = _("Not syncing, username or password unset.")
            if interactive:
                ui.utils.showWarning(msg)
            return
        if self.deck is None:
            # qt on linux incorrectly accepts shortcuts for disabled actions
            return
        if not self.deck.syncName and not onlyMerge:
            if not self.onSaveAs():
                return
        # bug triggered by preferences dialog - underlying c++ widgets are not
        # garbage collected until the middle of the child thread
        import gc; gc.collect()
        self.syncThread = ui.sync.Sync(self.deck, u, p, interactive, create, onlyMerge)
        self.connect(self.syncThread, SIGNAL("setStatus"), self.setSyncStatus)
        self.connect(self.syncThread, SIGNAL("showWarning"), ui.utils.showWarning)
        self.connect(self.syncThread, SIGNAL("moveToState"), self.moveToState)
        self.connect(self.syncThread, SIGNAL("noMatchingDeck"), self.selectSyncDeck)
        self.connect(self.syncThread, SIGNAL("cleanNewDeck"), self.cleanNewDeck)
        self.syncThread.start()
        self.setEnabled(False)
        while not self.syncThread.isFinished():
            self.app.processEvents()
            self.syncThread.wait(100)
        # the priorities on the server are different (no phone, etc)
        self.deck.updatePriorities()
        self.setEnabled(True)
        return self.syncThread.ok

    def selectSyncDeck(self, decks, create=True):
        name = ui.sync.DeckChooser(self, self.deck, decks, create).getName()
        if name:
            if name == self.deck.syncName:
                self.syncDeck(create=True)
            else:
                self.deck.syncName = name
                self.syncDeck()
        else:
            if not create:
                # called via 'new' - close
                self.cleanNewDeck()

    def cleanNewDeck(self):
        "Unload a new deck if an initial sync failed."
        self.deck = None
        self.moveToState("initial")

    def setSyncStatus(self, text, *args):
        self.setStatus(text, *args)
        self.mainWin.mainText.clear()
        self.mainWin.mainText.setHtml("<h1>" + text + "</h1>")

    # Menu, title bar & status
    ##########################################################################

    deckRelatedMenuItems = (
        "Save",
        "SaveAs",
        "Close",
        "Addcards",
        "Editdeck",
        "Syncdeck",
        "DisplayProperties",
        "DeckProperties",
        "ModelProperties",
        "Export",
        "MarkCard",
        "Graphs",
        "Dstats",
        "Kstats",
        "Cstats",
        )

    deckRelatedMenus = (
        "Tools",
        )

    def connectMenuActions(self):
        self.connect(self.mainWin.actionNew, SIGNAL("triggered()"), self.onNew)
        self.connect(self.mainWin.actionOpen, SIGNAL("triggered()"), self.onOpen)
        self.connect(self.mainWin.actionOpenSamples, SIGNAL("triggered()"), self.onOpenSamples)
        self.connect(self.mainWin.actionSave, SIGNAL("triggered()"), self.onSave)
        self.connect(self.mainWin.actionSaveAs, SIGNAL("triggered()"), self.onSaveAs)
        self.connect(self.mainWin.actionClose, SIGNAL("triggered()"), self.saveAndClose)
        self.connect(self.mainWin.actionExit, SIGNAL("triggered()"), self, SLOT("close()"))
        self.connect(self.mainWin.actionSyncdeck, SIGNAL("triggered()"), self.syncDeck)
        self.connect(self.mainWin.actionDeckProperties, SIGNAL("triggered()"), self.onDeckProperties)
        self.connect(self.mainWin.actionDisplayProperties, SIGNAL("triggered()"),self.onDisplayProperties)
        self.connect(self.mainWin.actionAddcards, SIGNAL("triggered()"), self.onAddCard)
        self.connect(self.mainWin.actionEditdeck, SIGNAL("triggered()"), self.onEditDeck)
        self.connect(self.mainWin.actionPreferences, SIGNAL("triggered()"), self.onPrefs)
        self.connect(self.mainWin.actionLookup_es, SIGNAL("triggered()"), self.onLookupEdictSelection)
        self.connect(self.mainWin.actionLookup_esk, SIGNAL("triggered()"), self.onLookupEdictKanjiSelection)
        self.connect(self.mainWin.actionLookup_expr, SIGNAL("triggered()"), self.onLookupExpression)
        self.connect(self.mainWin.actionLookup_mean, SIGNAL("triggered()"), self.onLookupMeaning)
        self.connect(self.mainWin.actionLookup_as, SIGNAL("triggered()"), self.onLookupAlcSelection)
        self.connect(self.mainWin.actionDstats, SIGNAL("triggered()"), self.onDeckStats)
        self.connect(self.mainWin.actionKstats, SIGNAL("triggered()"), self.onKanjiStats)
        self.connect(self.mainWin.actionCstats, SIGNAL("triggered()"), self.onCardStats)
        self.connect(self.mainWin.actionGraphs, SIGNAL("triggered()"), self.onShowGraph)
        self.connect(self.mainWin.actionAbout, SIGNAL("triggered()"), self.onAbout)
        self.connect(self.mainWin.actionReportbug, SIGNAL("triggered()"), self.onReportBug)
        self.connect(self.mainWin.actionForum, SIGNAL("triggered()"), self.onForum)
        self.connect(self.mainWin.actionStarthere, SIGNAL("triggered()"), self.onStartHere)
        self.connect(self.mainWin.actionImport, SIGNAL("triggered()"), self.onImport)
        self.connect(self.mainWin.actionExport, SIGNAL("triggered()"), self.onExport)
        self.connect(self.mainWin.actionMarkCard, SIGNAL("toggled(bool)"), self.onMark)
        self.connect(self.mainWin.actionSuspendCard, SIGNAL("triggered()"), self.onSuspend)
        self.connect(self.mainWin.actionModelProperties, SIGNAL("triggered()"), self.onModelProperties)
        self.connect(self.mainWin.actionRepeatQuestionAudio, SIGNAL("triggered()"), self.onRepeatQuestion)
        self.connect(self.mainWin.actionRepeatAnswerAudio, SIGNAL("triggered()"), self.onRepeatAnswer)
        self.connect(self.mainWin.actionRepeatAudio, SIGNAL("triggered()"), self.onRepeatAudio)
        self.connect(self.mainWin.actionUndoAnswer, SIGNAL("triggered()"), self.onUndoAnswer)

    def enableDeckMenuItems(self, enabled=True):
        "setEnabled deck-related items."
        for item in self.deckRelatedMenus:
            getattr(self.mainWin, "menu" + item).setEnabled(enabled)
        for item in self.deckRelatedMenuItems:
            getattr(self.mainWin, "action" + item).setEnabled(enabled)
        if self.deckIsReadOnly():
            self.mainWin.actionSave.setEnabled(False)

    def disableDeckMenuItems(self):
        "Disable deck-related items."
        self.enableDeckMenuItems(enabled=False)

    def updateTitleBar(self):
        "Display the current deck and card count in the titlebar."
        title=ankiqt.appName + " " + ankiqt.appVersion
        if self.deck != None:
            deckpath = self.deck.name()
            if self.deck.modifiedSinceSave():
                deckpath += "*"
            title = _("%(path)s (%(facts)d facts, %(cards)d cards)"
                      " - %(title)s") % {
                # FIXME: safe to assume filesystem is utf-8?
                "path": deckpath,
                "title": title,
                "cards": len(self.deck),
                "facts": len(self.deck.facts),
                }

        self.setWindowTitle(title)

    def setStatus(self, text, timeout=3000):
        self.mainWin.statusbar.showMessage(text, timeout)

    def onStartHere(self):
        QDesktopServices.openUrl(QUrl(ankiqt.appHelpSite))

    def alterShortcuts(self):
        if sys.platform.startswith("darwin"):
            self.mainWin.actionAddcards.setShortcut(_("Ctrl+D"))

    def updateRecentFilesMenu(self):
        files = []
        # work around an old bug
        for f in self.config['recentDeckPaths']:
            if f not in files:
                files.append(f)
        if not files:
            self.mainWin.menuOpenRecent.setEnabled(False)
        else:
            self.mainWin.menuOpenRecent.setEnabled(True)
        self.mainWin.menuOpenRecent.clear()
        n = 1
        for file in files:
            a = QAction(self)
            a.setShortcut(_("Alt+%d" % n))
            a.setText(os.path.basename(file))
            self.connect(a, SIGNAL("triggered()"),
                         lambda n=n: self.loadRecent(n-1))
            self.mainWin.menuOpenRecent.addAction(a)
            n += 1

    def updateMarkAction(self):
        self.mainWin.actionMarkCard.blockSignals(True)
        if self.currentCard.hasTag("Marked"):
            self.mainWin.actionMarkCard.setChecked(True)
        else:
            self.mainWin.actionMarkCard.setChecked(False)
        self.mainWin.actionMarkCard.blockSignals(False)

    def disableCardMenuItems(self):
        if not self.lastCard:
            self.mainWin.actionUndoAnswer.setEnabled(False)
        self.mainWin.actionMarkCard.setEnabled(False)
        self.mainWin.actionSuspendCard.setEnabled(False)
        self.mainWin.actionRepeatQuestionAudio.setEnabled(False)
        self.mainWin.actionRepeatAnswerAudio.setEnabled(False)
        self.mainWin.actionRepeatAudio.setEnabled(False)

    def enableCardMenuItems(self):
        self.mainWin.actionUndoAnswer.setEnabled(not not self.lastCard)
        self.mainWin.actionMarkCard.setEnabled(True)
        self.mainWin.actionSuspendCard.setEnabled(True)
        self.mainWin.actionRepeatQuestionAudio.setEnabled(
            hasSound(self.currentCard.question))
        self.mainWin.actionRepeatAnswerAudio.setEnabled(
            hasSound(self.currentCard.answer) and self.state != "getQuestion")
        self.mainWin.actionRepeatAudio.setEnabled(
            self.mainWin.actionRepeatQuestionAudio.isEnabled() or
            self.mainWin.actionRepeatAnswerAudio.isEnabled())

    # Auto update
    ##########################################################################

    def setupAutoUpdate(self):
        self.autoUpdate = ui.update.LatestVersionFinder(self)
        self.connect(self.autoUpdate, SIGNAL("newVerAvail"), self.newVerAvail)
        self.autoUpdate.start()

    def newVerAvail(self, version):
        ui.update.askAndUpdate(self, version)

    # User customisations
    ##########################################################################

    def loadUserCustomisations(self):
        # look for config file
        dir = self.config.configPath
        file = os.path.join(dir, "custom.py")
        plugdir = os.path.join(dir, "plugins")
        sys.path.insert(0, dir)
        if os.path.exists(file):
            try:
                import custom
            except:
                print "Error in custom.py"
                print traceback.print_exc()
        sys.path.insert(0, plugdir)
        import glob
        plugins = [f.replace(".py", "") for f in os.listdir(plugdir) \
                   if f.endswith(".py")]
        plugins.sort()
        for plugin in plugins:
            try:
                __import__(plugin)
            except:
                print "Error in %s.py" % plugin
                print traceback.print_exc()

    # Locking
    ##########################################################################

    def checkLock(self):
        lockFile = os.path.join(self.config.configPath, "lock")
        if os.path.exists(lockFile):
            ret = QMessageBox.warning(None, "Anki", _(
                "An Anki lock file exists, which indicates that\n"
                "another copy of Anki is running. If Anki or your\n"
                "computer crashed last time, you can ignore this.\n\n"
                "Would you like to ignore the lock file and start\n"
                "Anki anyway?"),
                                QMessageBox.Yes | QMessageBox.No,
                                QMessageBox.No)
            if ret == QMessageBox.Yes:
                return
            else:
                # FIXME: is this the best approach?
                sys.exit(1)
        else:
            open(lockFile, "w")

    def removeLock(self):
        lockFile = os.path.join(self.config.configPath, "lock")
        try:
            os.unlink(lockFile)
        except OSError:
            # may have been removed by another instance
            pass

    # Mac support
    ##########################################################################

    def macHacks(self):
        if sys.platform.startswith("darwin"):
            self.setUnifiedTitleAndToolBarOnMac(True)
            qt_mac_set_menubar_icons(False)

    # Font localisation
    ##########################################################################

    def setupFonts(self):
        for (s, p) in anki.fonts.substitutions():
            QFont.insertSubstitution(s, p)

    # Sounds
    ##########################################################################

    def onRepeatQuestion(self):
        playFromText(self.currentCard.question)

    def onRepeatAnswer(self):
        playFromText(self.currentCard.answer)

    def onRepeatAudio(self):
        if self.state == "showQuestion":
            playFromText(self.currentCard.question)
        else:
            if hasSound(self.currentCard.answer):
                playFromText(self.currentCard.answer)
            else:
                playFromText(self.currentCard.question)
