# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

from PyQt4.QtGui import *
from PyQt4.QtCore import *
import os, types, socket
import ankiqt
import anki
from anki.sync import SyncClient, HttpSyncServerProxy
from anki.errors import *
import ankiqt.forms

# Synchronising a deck with a public server
##########################################################################

class Sync(QThread):

    def __init__(self, deck, user, pwd, interactive, create, onlyMerge):
        QThread.__init__(self)
        self.deck = deck
        self.interactive = interactive
        self.user = user
        self.pwd = pwd
        self.create = create
        self.ok = True
        self.onlyMerge = onlyMerge

    def setStatus(self, msg, timeout=5000):
        self.emit(SIGNAL("setStatus"), msg, timeout)

    def run(self):
        self.syncDeck()

    def error(self, error):
        self.setStatus(self.getErrorMessage(error))
        if self.onlyMerge:
            # new file needs cleaning up
            self.emit(SIGNAL("cleanNewDeck"))
        else:
            self.emit(SIGNAL("moveToState"), "initial")

    def getErrorMessage(self, error):
        if error.data.get('status') == "invalidUser":
            msg=_("Please double-check your username.")
        elif error.data.get('status') == "invalidPass":
            msg=_("Please double-check your password.")
        elif error.data.get('status') == "oldVersion":
            msg=_("The sync protocol has changed. Please upgrade.")
        elif error.data.get('type') == "noResponse":
            msg=_("Server is down or operation failed.")
        else:
            msg=_("Unknown error: %s") % `error.data`
        return msg

    def errorInProgress(self, error, lastOp):
        msg = _("Syncing failed halfway through. If it fails again, please\n"
                "email your deck to the Anki author.\n\n")
        msg += self.getErrorMessage(error)
        msg += _("\nThe last operation was '%s'.") % lastOp
        self.setStatus(_("Syncing failed."))
        self.emit(SIGNAL("showWarning"), msg)

    def syncDeck(self):
        self.setStatus(_("Syncing.."), 0)
        # connect, check auth
        proxy = HttpSyncServerProxy(self.user, self.pwd)
        try:
            proxy.connect(ankiqt.appVersion)
        except SyncError, e:
            return self.error(e)
        # exists on server?
        if not proxy.hasDeck(self.deck.syncName):
            if self.create:
                try:
                    proxy.createDeck(self.deck.syncName)
                except SyncError, e:
                    return self.error(e)
            else:
                self.emit(SIGNAL("noMatchingDeck"), proxy.decks.keys(),
                          not self.onlyMerge)
                self.setStatus("")
                self.emit(SIGNAL("moveToState"), "initial")
                return
        # need to do anything?
        proxy.deckName = self.deck.syncName
        if self.deck.modified == proxy.modified():
            self.setStatus(_("Sync: nothing to do"))
            self.emit(SIGNAL("moveToState"), "initial")
            return
        # start syncing
        client = SyncClient(self.deck)
        client.setServer(proxy)
        # handle the sync manually so we can give more feedback
        client.localTime = client.modified()
        client.remoteTime = proxy.modified()
        lastOp = None

        try:
            self.setStatus(_("Fetching summary from server.."), 0)
            lastOp = "summary"
            client.sum = proxy.summary()
            self.setStatus(_("Syncing models.."), 0)
            lastOp = "models"
            client.applyDiff("models", client.diffModels(client.sum['models']))
            self.setStatus(_("Syncing facts.."), 0)
            lastOp = "facts"
            client.applyDiff("facts", client.diffFacts(client.sum['facts']))
            self.setStatus(_("Syncing cards.."), 0)
            lastOp = "cards"
            client.applyDiff("cards", client.diffCards(client.sum['cards']))
            self.setStatus(_("Syncing deck scheduling data.."), 0)
            lastOp = "sched"
            client.oneWay("updateSched", "bundleSched")
            self.setStatus(_("Syncing facts metadata.."), 0)
            lastOp = "factsmeta"
            client.oneWay("updateFactsMeta", "bundleFactsMeta")
            self.setStatus(_("Syncing deck metadata.."), 0)
            lastOp = "deck"
            client.oneWay("updateDeck", "bundleDeck")
            proxy.saveDeck()
        except SyncError, e:
            self.errorInProgress(e, lastOp)
            self.emit(SIGNAL("moveToState"), "initial")
            self.ok = False
            return
        self.setStatus(_("Syncing..done. Sent %(a)d/%(b)d, "
                         "received %(c)d/%(d)d cards.") % {
                "a": len(client.diffs['cards'][0]),
                "b": len(client.diffs['cards'][1]),
                "c": len(client.diffs['cards'][2]),
                "d": len(client.diffs['cards'][3])})
        self.emit(SIGNAL("moveToState"), "initial")

# Choosing a deck to sync to
##########################################################################

class DeckChooser(QDialog):

    def __init__(self, parent, deck, decks, create):
        QDialog.__init__(self, parent)
        self.parent = parent
        self.decks = decks
        self.deck = deck
        self.dialog = ankiqt.forms.syncdeck.Ui_DeckChooser()
        self.dialog.setupUi(self)
        self.create = create
        if self.create:
            self.dialog.decks.addItem(QListWidgetItem(
                _("Create '%s' on server") % self.deck.syncName))
        self.decks.sort()
        for name in decks:
            name = os.path.splitext(name)[0]
            item = QListWidgetItem(_("Merge with '%s' on server") % name)
            self.dialog.decks.addItem(item)
        self.dialog.decks.setCurrentRow(0)
        # the list widget will swallow the enter key
        s = QShortcut(QKeySequence("Return"), self)
        self.connect(s, SIGNAL("activated()"), self.accept)
        self.name = None

    def getName(self):
        self.exec_()
        return self.name

    def accept(self):
        idx = self.dialog.decks.currentRow()
        if self.create:
            offset = 1
        else:
            offset = 0
        if idx == 0 and self.create:
            self.name = self.deck.syncName
        else:
            self.name = self.decks[self.dialog.decks.currentRow() - offset]
        self.close()

