#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

"""\
Card Models - control the way cards are generated and handled
==============================================================
"""

import time, re
from anki.cards import Card
from anki.lang import _
from anki.ids import *
from anki.fonts import toPlatform

class CardModels(IDList):
    def __unicode__(self):
        return u"<CardModels size=%d>" % len(self)

class CardModel(IDObject):
    """Represents how to generate the front and back of a card."""
    def __init__(self, name="", description="",
                 qformat="", aformat=""):
        IDObject.__init__(self)
        self.name = name
        self.description = description
        # question/answer format
        self.qformat = qformat
        self.aformat = aformat
        # last card format
        self.lformat = None
        # question/answer editor format (not used yet)
        self.qedformat = None
        self.aedformat = None
        self.questionInAnswer = False
        self.delay = 86400
        # non-empty fields required to generate this card
        self.required = []
        self.display = {
            "question": {
            "fontFamily": "Arial",
            "fontSize": 20,
            "fontColour": "#000000",
            "align": 0,
            },
            "answer": {
            "fontFamily": "Arial",
            "fontSize": 20,
            "fontColour": "#000000",
            "align": 0,
            },
            "last": {
            "fontFamily": "Arial",
            "fontSize": 20,
            "fontColour": "#000000",
            },
            "editQuestion": {
            "fontFamily": None,
            "fontSize": None,
            },
            "editAnswer": {
            "fontFamily": None,
            "fontSize": None,
            },
            }

    def makeCard(self, fact):
        "Return a card for FACT based on card MODEL."
        c = Card(fact, self)
        return c

    def timeLeft(self, card):
        "Time until card can be displayed. May be 0."
        if not card.fact.lastCard or card.fact.lastCard == card.model:
            return 0
        return card.fact.lastAnswered + self.delay - time.time()

    def renderQuestion(self, card, fact, fieldName="question", type="text"):
        "Return a formatted question for FACT."
        if fieldName == "question":
            field = self.qformat
        elif fieldName == "answer":
            field = self.aformat
        # update the fields to include formatting
        fields = {}
        fields.update(fact)
        fields['tags'] = ", ".join(card.tags + fact.tags + [card.model.name])
        textfields = {}
        textfields.update(fields)
        for (k, v) in fields.items():
            # generate pure text entries
            fields["text:"+k] = v
            textfields["text:"+k] = v
            if v:
                # convert newlines to html
                v = v.replace("\n", "<br>")
                fields[k] = '<span class="%s">%s</span>' % (k.replace(" ",""), v)
        try:
            val = field % fields
            txtval = field % textfields
        except (KeyError, TypeError, ValueError):
            return _("[invalid format; see model properties]")
        if not txtval:
            return _("[empty]")
        if type == "html":
            val = '<div class="%s">%s</div>' % (fieldName, val)
            # we need to do this with tables due to qt's poor html handling
            if self.display[fieldName]['align'] == 0:
                align = "center"
            elif self.display[fieldName]['align'] == 1:
                align = "left"
            else:
                align = "right"
            val = (("<center><table width=95%%><tr><td align=%s>" % align) +
                   val +
                   "</td></tr></table></center>")
            return val
        else:
            return txtval

    def renderAnswer(self, card, fact, type="text"):
        "Return a formatted answer for FACT."
        return self.renderQuestion(card, fact, "answer", type=type)

    def css(self):
        "Return the CSS markup for this card."
        t = ""
        for type in ("question", "answer"):
            t += ".%s { font-family: \"%s\"; color: %s; font-size: %dpx; }\n" % (
                type,
                toPlatform(self.display[type]["fontFamily"]),
                self.display[type]["fontColour"],
                self.display[type]["fontSize"])
        return t

    def setModified(self, newTime=None):
        if not newTime:
            newTime = time.time()
        self.__dict__['modified'] = newTime
        if self.__dict__.get('parent', None) is not None:
            self.parent.setModified(newTime)

    def __unicode__(self):
        return "<CardModel n='%s' q='%s' a='%s'>" % (self.name,
                                                     self.qformat,
                                                     self.aformat)
    def __repr__(self):
        return self.__unicode__()
