#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

"""\
Deck related
====================
"""
__docformat__ = 'restructuredtext'

import cPickle, cStringIO, time, os, shutil, random, sys, re, tempfile
from anki.cards import Card
from anki.facts import Fact, Facts
from anki.lang import _
from anki.errors import *
from anki.ids import *

latexPreamble = ("\\documentclass[12pt]{article}\n"
                 "\\pagestyle{empty}\n"
                 "\\begin{document}")
latexPostamble = "\\end{document}"
latexDviPngCmd = "dvipng -D 200 -T tight"

class Deck(IDList):
    "A list of cards, with a facts database and scheduling manager."

    def __init__(self, sched, model):
        "Create a new deck with SCHED and default fact MODEL."
        IDList.__init__(self, None)
        self.deckFormat = "anki"
        self.deckVersion = 0
        self.lastLoaded = time.time()
        self.sched = sched
        self.sched.deck = self
        self.facts = Facts(self)
        self.facts.models.setModel(model)
        # no sync name indicates syncing is disabled
        self.syncName = None
        self.description = ""
        self.path = None
        # not yet used
        self.undoList = []
        self.userVars = {}
        # latex
        self.latexPreamble = latexPreamble
        self.latexPostamble = latexPostamble
        self.latexDviPngCmd = latexDviPngCmd

    def name(self):
        if self.path:
            return os.path.splitext(os.path.basename(self.path))[0]
        return _("Untitled")

    def modifiedSinceSave(self):
        return self.modified > self.lastLoaded

    def __unicode__(self):
        return u"<deck name='%s' syncName='%s' cards=%d facts=%d>" % (
            self.name(), self.syncName, len(self), len(self.facts))

    # Facts
    ##########################################################################

    def addFact(self, fact):
        self.facts.addFact(fact)
        if not fact.model:
            fact.model = self.facts.models.current
        cards = fact.model.makeCards(fact)
        self.addCards(cards, fact.model.position)
        return cards

    def removeFactIfUnused(self, fact):
        "Remove the fact if there are no cards. Don't update modtime."
        if not fact.cards:
            try:
                self.facts.remove(fact)
            except ValueError:
                # ignore failure - the fact was probably removed in the
                # syncing code
                pass

    def factIsInvalid(self, fact):
        "True if existing fact is invalid. Returns the error."
        try:
            fact.model.ensureValid(fact)
            fact.model.ensureUnique(fact, self.facts)
        except FactInvalidError, e:
            return e
        return False

    # Cards
    ##########################################################################

    def addCards(self, cards, position):
        "Add cards to the list of new cards."
        if position:
            # append to end of deck
            idx = len(self)
        else:
            # random position
            idx = random.randrange(0, 1 + len(self))
        n = 0
        for card in cards:
            self.addCard(card, idx+n)
            n += 1

    def addCard(self, card, index):
        self.sched.addToCard(card)
        self.insert(index, card)

    def __delitem__(self, index):
        fact = self[index].fact
        try:
            fact.cards.remove(self[index])
        except ValueError:
            sys.stderr.write(
                "Warning: fact's card list didn't match. Ignoring.\n")
        IDList.__delitem__(self, index)
        self.removeFactIfUnused(fact)

    def remove(self, card):
        self.__delitem__(self.index(card))

    # Fields
    ##########################################################################

    def fieldUseCount(self, model, fieldName):
        "Number of non-empty occurances of FIELD across the deck."
        count = 0
        for card in self:
            if model == card.fact.model and card.fact.get(fieldName):
                count += 1
        return count

    def removeField(self, model, field):
        self.facts.removeField(model, field)
        if field.name in model.unique:
            model.unique.remove(field.name)
        if field.name in model.required:
            model.required.remove(field.name)
        model.setModified()

    # Models
    ##########################################################################

    def removeModel(self, model):
        "Remove MODEL from deck and remove any referencing cards/facts."
        cards = [c for c in self if c.fact.model == model]
        for card in cards:
            self.remove(card)
        self.facts.removeModel(model)

    def modelUseCount(self, model):
        "Return number of cards using model."
        count = 0
        for card in self:
            if card.fact.model == model:
                count += 1
        return count

    # Card models
    ##########################################################################

    def cardModelUseCount(self, cardModel):
        "Return the number of cards using cardModel."
        count = 0
        for card in self:
            if card.model == cardModel:
                count += 1
        return count

    def removeCardModel(self, model, cardModel):
        "Remove all cards that use CARDMODEL from the deck."
        cards = [c for c in self if c.model == cardModel]
        for card in cards:
            self.remove(card)
        model.removeCardModel(cardModel)

    # Tags
    ##########################################################################

    def allTags(self):
        "Return a hash listing tags in model, fact and cards."
        tags = {'model': [],
                'card': [],
                'fact': [],
                'cardModel': []}
        for model in self.facts.models:
            for tag in model.tags:
                if tag not in tags['model']:
                    tags['model'].append(tag)
        for card in self:
            for tag in card.tags:
                if tag not in tags['card']:
                    tags['card'].append(tag)
            if card.model.name not in tags['cardModel']:
                tags['cardModel'].append(card.model.name)
        for fact in self.facts:
            for tag in fact.tags:
                if tag not in tags['fact']:
                    tags['fact'].append(tag)
        tags['all'] = list(set(tags['model'] +
                               tags['card'] +
                               tags['fact'] +
                               tags['cardModel']))
        return tags

    def updatePriorities(self):
        for card in self:
            card.updatePriority()

    # Media
    ##########################################################################

    def mediaDir(self, create=True):
        "Return the media directory, creating if necessary. None if no path."
        if not self.path:
            return None
        dir = re.sub("(?i)\.(anki|fc)$", ".media", self.path)
        if not os.path.exists(dir) and create:
            try:
                os.mkdir(dir)
            except OSError:
                # permission denied
                return None
        return dir

    def addMedia(self, path):
        """Add PATH to the media directory.
Rename for uniqueness if not the same. Return the new path."""
        file = os.path.basename(path)
        location = os.path.join(self.mediaDir(), file)
        location = location.replace('"', "")
        if os.path.exists(location):
            print "uniqueness check NYI"
        else:
            shutil.copy2(path, location)
        return os.path.basename(location)

    def renameMediaDir(self, oldPath):
        "Copy oldPAth to our current media dir. "
        assert os.path.exists(oldPath)
        newPath = self.mediaDir(create=False)
        if os.path.exists(newPath):
            return
        shutil.copytree(oldPath, newPath)

# Deck storage and generation
##########################################################################

class DeckStorage(object):

    numBackups = 15

    def load(self, path, backup=True):
        "Load PATH as a native file."
        deck = self.fromFile(path)
        if getattr(deck, 'deckFormat', "fc") == "fc":
            raise DeckWrongFormatError(
                info="Old deck. Please import instead of loading.")
        deck.lastLoaded = time.time()
        # upgrade scheduler first
        deck.sched.afterLoad()
        self.upgrade(deck)
        if backup:
            self.backup(deck.path)
        return deck

    def fromFile(self, path, dummyClasses=False):
        "Load a deck from PATH. Version is not checked."
        assert isinstance(path, unicode)
        self.dummyClasses = dummyClasses
        file = open(path, "rb")
        try:
            try:
                data = file.read()
            except (IOError, OSError):
                raise ImportError
        finally:
            file.close()
        unpickler = cPickle.Unpickler(cStringIO.StringIO(data))
        unpickler.find_global = self.transformClasses
        deck = unpickler.load()
        deck.path = os.path.abspath(path)
        return deck

    def transformClasses(self, m, c):
        "Handle pre 0.3 decks."
        class EmptyClass(object):
            def resetSessionStats(self):
                pass
        if m == "anki.card":
            return EmptyClass
        if c == "Manager":
            return EmptyClass
        if self.dummyClasses:
            # use empty classes to retrieve old data
            if c == "Card":
                return EmptyClass
        if m == "fc":
            raise DeckWrongFormatError(
                info="Deck is too old. Please upgrade with 0.2.8 first.")
        # handle obsolete paths
        if c == "Deck":
            m = "anki.deck"
        if c == "Card":
            m = "anki.cards"
        return getattr(__import__(m, None, None, [m]), c)

    def save(self, deck, path=None, backup=True):
        "Save the deck to a file. PATH should be unicode."
        oldMediaDir = deck.mediaDir()
        if path:
            deck.path = path
        else:
            if not deck.path:
                raise ValueError("No path provided to save.")
        assert isinstance(deck.path, unicode)
        if backup:
            # this will raise an exception if something goes wrong
            self.backup(deck.path)
        if not deck.syncName:
            deck.syncName = deck.name()
        # write to a temp file
        (fd, tmpname) = tempfile.mkstemp(dir=os.path.dirname(deck.path),
                                      suffix=".anki")
        tmpfile = os.fdopen(fd, 'wb')
        cPickle.dump(deck, tmpfile, 2)
        tmpfile.close()
        # the write was successful, delete old deck (if exists) and rename
        if os.path.exists(deck.path):
            os.unlink(deck.path)
        os.rename(tmpname, deck.path)
        # handle media directory
        if oldMediaDir:
            deck.renameMediaDir(oldMediaDir)
        deck.lastLoaded = time.time()

    def backup(self, path):
        "Return False on failure."
        # can't back it up if it didn't exist before
        if not os.path.exists(path):
            return
        from anki import Controller
        # make sure the backup dir exists
        if not os.path.exists(Controller.backupDir):
            os.makedirs(Controller.backupDir)
        # if the files are identical, don't make a new backup
        if os.path.exists(self.backupName(path, 0)):
            backup = open(self.backupName(path, 0), "rb").read()
            current = open(path, "rb").read()
            if backup == current:
                return
        # remove the oldest backup if it exists
        oldest = self.backupName(path, self.numBackups)
        if os.path.exists(oldest):
            os.chmod(oldest, 0666)
            os.unlink(oldest)
        # move all the other backups up one
        for n in range(self.numBackups - 1, -1, -1):
            name = self.backupName(path, n)
            if os.path.exists(name):
                newname = self.backupName(path, n+1)
                if os.path.exists(newname):
                    os.chmod(newname, 0666)
                    os.unlink(newname)
                os.rename(name, newname)
        # save the current path
        newpath = self.backupName(path, 0)
        if os.path.exists(newpath):
            os.chmod(newpath, 0666)
            os.unlink(newpath)
        shutil.copy2(path, newpath)

    def backupName(self, path, num):
        from anki import Controller
        path = os.path.abspath(path)
        path = path.replace("\\", "!")
        path = path.replace("/", "!")
        path = path.replace(":", "")
        path = os.path.join(Controller.backupDir, path)
        path += ".backup-%d" % num
        return path

    def upgrade(self, deck):
        "Add any missing properties."
        # latex properties added in 0.3.7
        deck.latexPreamble = latexPreamble
        deck.latexPostamble = latexPostamble
        deck.latexDviPngCmd = latexDviPngCmd
