#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

"""\
Decorators - implementation specific support
===============================================================================

Decorators allow the deck to define specific features that are required, but
that can be resolved in real time. This includes things like automatic reading
generation, language-specific dictionary entries, etc.

A decorator has a list of tags which it answers to. For example, you may have
one decorator named 'Japanese dictionaries', which is activated whenever a
deck contains the decorator tag 'Japanese'.
"""

from anki.lang import _
from anki.errors import *
from anki.utils import findTag

class Decorator(object):

    def __init__(self, tags=None, name="", description=""):
        if not tags:
            tags = []
        self.tags = tags
        self.name = name
        self.description = description

    def onSubmit(self, fact):
        "Apply any last-minute modifications to FACT before addition."
        pass

    def onKeypress(self, fact):
        "Apply any changes to fact as it's being edited for the first time."
        pass

    def lookupItems(self):
        "Return a list of tuples."
        pass

    def doCmd(self, cmd, *args):
        "Run CMD."
        attr = getattr(self, cmd, None)
        if attr:
            attr(*args)

class Decorators(object):

    decorators = {}

    def add(decorator):
        "Add a decorator."
        Decorators.decorators[decorator.name] = decorator
    add = staticmethod(add)

    def doAll(cmd, dlist, *args):
        "Run CMD on all matching decorators in DLIST."
        for (name, decorator) in Decorators.decorators.items():
            for tag in dlist:
                if findTag(tag, decorator.tags):
                    decorator.doCmd(cmd, *args)
                    break
    doAll = staticmethod(doAll)

# Add bundled
from anki.support import japanese; japanese.install()
from anki.support import chinese; chinese.install()
