#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

"""\
Exporting support
==============================
"""
__docformat__ = 'restructuredtext'

import copy
from anki import Controller
from anki.cards import Card
from anki.sync import SyncTools
from anki.lang import _
import anki.json2 as json

class Exporter(object):
    def __init__(self, deck):
        self.deck = deck
        self.limitTags = []

    def exportInto(self, path):
        file = open(path, "wb")
        self.doExport(file)
        file.close()

    def escapeText(self, text):
        "Escape newlines and tabs."
        text = text.replace("\n", "<br>")
        text = text.replace("\t", " " * 8)
        return text

    def allCards(self):
        "Return all cards, limited by tags."
        cards = [c for c in self.deck if self.hasTags(c)]
        self.count = len(cards)
        return cards

    def allFacts(self):
        "Return all facts, limited by tags."
        facts = [f for f in self.deck.facts if self.hasTags(f)]
        self.count = len(facts)
        return facts

    def hasTags(self, obj):
        if not self.limitTags:
            return True
        for tag in self.limitTags:
            if obj.hasTag(tag):
                return True
        return False

class TextCardExporter(Exporter):

    key = _("Text files (*.txt)")
    ext = ".txt"

    def __init__(self, deck):
        Exporter.__init__(self, deck)
        self.includeSchedulingInfo = True

    def doExport(self, file):
        out = "\n".join(["%s\t%s%s" % (self.escapeText(c.question),
                                       self.escapeText(c.answer),
                                       self.schedulingInfo(c))
                          for c in self.allCards()])
        if out:
            out += "\n"
        file.write(out.encode("utf-8"))

    def schedulingInfo(self, card):
        if not self.includeSchedulingInfo:
            return ""
        d = card.__dict__.copy()
        for k in "fact", "model", "history", "parent", "deviation":
            if k in d:
                del d[k]
        d['fid'] = card.fact.id
        d['ftags'] = card.fact.tags
        d['mtags'] = card.fact.model.tags
        d['cmname'] = card.model.name
        d['cmid'] = card.model.id
        d['mid'] = card.fact.model.id
        return "\n" + json.write(d)

class TextFactExporter(Exporter):

    key = _("Text files (*.txt)")
    ext = ".txt"

    def doExport(self, file):
        out = "\n".join(["\t".join([self.escapeText(v) for v in f.values()])
                         for f in self.allFacts()])
        if out:
            out += "\n"
        file.write(out.encode("utf-8"))

# FIXME: media support
class AnkiExporter(Exporter):

    key = _("Anki decks (*.anki)")
    ext = ".anki"

    def __init__(self, deck):
        Exporter.__init__(self, deck)
        self.includeSchedulingInfo = False

    def exportInto(self, path):
        self.newDeck = Controller.Deck()
        for card in self.allCards():
            self.addModel(card.fact.model)
            fact = self.addFact(card.fact)
            newCard = Card()
            newCard.__dict__.update(card.__dict__)
            # prevent the card change from updating the original deck
            del newCard.parent
            newCard.fact = fact
            if not self.includeSchedulingInfo:
                self.newDeck.sched.resetCard(newCard)
            self.newDeck.addCard(newCard, len(self.newDeck))
        # delete lastTags
        self.newDeck.facts.lastTags = []
        # save
        ds = Controller.DeckStorage()
        ds.save(self.newDeck, path)

    def addModel(self, model):
        "Add model to new deck if not already there."
        if model not in self.newDeck.facts.models:
            self.newDeck.facts.models.addModel(model)
        return model

    def addFact(self, fact):
        "Add fact to new deck if not already there."
        # fact is changed when adding, so prevent updates to old deck
        newfact = copy.copy(fact)
        del newfact.parent
        if newfact not in self.newDeck.facts:
            self.newDeck.facts.addFact(newfact, validate=False)
        return fact

# Export modules
##########################################################################

Exporters = (
    (_("Anki deck (*.anki)"), AnkiExporter),
    (_("Cards in tab-separated text file (*.txt)"), TextCardExporter),
    (_("Facts in tab-separated text file (*.txt)"), TextFactExporter),
    )
