#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

"""\
Fact lists - a list of facts which are used to generate cards.
==============================================================
"""
__docformat__ = 'restructuredtext'

import time
from anki.cards import *
from anki.errors import *
from anki.models import Models
from anki.ids import *

class Facts(IDList):
    "An order-independent list of facts used to generate cards."

    def __init__(self, parent=None):
        IDList.__init__(self, parent)
        self.models = Models(self)
        self.lastTags = []

    # Facts
    ##########################################################################

    def addFact(self, fact, validate=True):
        "Add FACT based on the current model."
        model = fact.model or self.models.current
        model.onSubmit(fact)
        if validate:
            model.ensureValid(fact)
            model.ensureUnique(fact, self)
        fact.model = model
        if not fact.id:
            fact.id = hash("".join(sorted(fact.values())))
        self.append(fact)
        self.lastTags = fact.tags[:]

    def __unicode__(self):
        s = u"<facts facts=%d models=%d curmodel='%s'>" % (
            len(self), len(self.models), getattr(self.models.current, 'name', None))
        return s

    # Models
    ##########################################################################

    def removeModel(self, model, deleteFacts=True):
        "Remove MODEL and delete all referencing facts."
        if deleteFacts:
            toDelete = [f for f in self if f.model == model]
            for fact in toDelete:
                self.remove(fact)
        self.models.remove(model)
        # if that was the current model, we need to select something else
        if self.models.current == model:
            if self.models:
                self.models.current = self.models[0]
            else:
                self.models.current = None
        self.setModified()

    def modelUseCount(self, model):
        "Number of facts using MODEL."
        count = 0
        for fact in self:
            if fact.model == model:
                count += 1
        return count

    # Fields
    ##########################################################################

    def addField(self, model, field):
        "Add FIELD to MODEL and update cards."
        model.addField(field)
        for fact in self:
            if fact.model == model:
                fact[field.name] = ""

    def removeField(self, model, field):
        "Remove FIELDNAME from MODEL and update facts."
        model.fields.remove(field)
        for fact in self:
            if fact.model == model:
                del fact[field.name]

    def renameField(self, model, field, newName):
        "Change FIELD's name in MODEL and update FIELD in all facts."
        for fact in self:
            if fact.model == model:
                assert newName not in fact
                fact[newName] = fact[field.name]
                del fact[field.name]
        # update model
        if field.name in model.required:
            model.required.remove(field.name)
            model.required.append(newName)
        if field.name in model.unique:
            model.unique.remove(field.name)
            model.unique.append(newName)
        # update card models
        for cm in model.allcards:
            cm.qformat = cm.qformat.replace("%%(%s)s" % field.name,
                                            "%%(%s)s" % newName)
            cm.aformat = cm.aformat.replace("%%(%s)s" % field.name,
                                            "%%(%s)s" % newName)
        field.name = newName
        field.setModified()

    def byField(self, field, value, substring=False):
        "A list of facts with FIELD = VALUE."
        results = []
        value = value.lower()
        for fact in self:
            data = fact.get(field, "").lower()
            if substring and value in data:
                results.append(fact)
            elif data == value:
                results.append(fact)
        return results

    def byTag(self, tag):
        "A list of facts with TAG."
        return [f for f in self if tag in f.tags]

    def allFields(self):
        "A list of all available field names."
        fs = []
        for model in self.models:
            for field in model.fields:
                fs.append(field.name)
        return list(set(fs))

class Fact(IDObject, dict):
    "A single fact."

    def __init__(self):
        IDObject.__init__(self)
        # start off with no id
        self.id = None
        self.tags = []
        self.model = None
        self.cards = []
        # prevent related cards from being displayed for a period
        self.lastAnswered = 0
        self.lastCard = None

    def cardAnswered(self, card):
        "Update log of last card."
        self.lastAnswered = time.time()
        self.lastCard = card.model

    def css(self):
        return "".join([f.css() for f in self.model.fields])

    def hasTag(self, tag):
        tag = tag.lower()
        return (tag in [t.lower() for t in self.tags] or
                tag in [t.lower() for t in self.model.tags])

    def __setattr__(self, k, v):
        self.__dict__[k] = v
        self.setModified()

    def __setitem__(self, k, v):
        dict.__setitem__(self, k, v)
        self.setModified()

    def __delitem__(self, i):
        dict.__delitem__(self, i)
        self.setModified()

    def __unicode__(self):
        return ", ".join(["%s: %s" % (k, v) for (k, v) in self.items()])
