#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

"""\
Importing support
==============================
"""
__docformat__ = 'restructuredtext'

import codecs, time, pickle, sys, re, os
import anki, anki.stdmodels
from anki.errors import *
from anki.models import *
from anki.facts import Fact
from anki.lang import _
from anki import Controller

# Common code
##########################################################################

class ForeignCard(list):
    "An temporary list of fields and metadata for a foreign card."

    def __init__(self):
        self.tags = []
        self.matchingTag = None
        # apply the default scheduling information. this may be updated
        # depending on the information available to us
        anki.Controller.scheduler().addToCard(self)

class ForeignFact(list):
    "A list of cards corresponding to one 'fact'."

    def __init__(self, *cards):
        self.extend(cards)

    def fields(self):
        return self[0]

class ForeignCardImporter(object):

    def __init__(self, deck, model, mapping):
        self.deck = deck
        self.model = model
        self.mapping = mapping
        self.total = 0
        self.log = []
        self.singles = 0
        self.dupes = 0
        self.multiples = 0
        self.missing = 0
        self.addedFacts = []
        self.autoSingle = False

    def importCards(self, cards):
        # preserve properties which are modified during import
        self.activatedCards = self.model.cards
        self.lastTags = self.deck.facts.lastTags
        self.deck.facts.models.setModel(self.model)
        for fact in self.factsFromCards(cards):
            self.addFact(fact)
        self.total = self.singles + self.multiples
        # restore the old list of activated cards
        self.model.cards = self.activatedCards
        self.deck.facts.lastTags = self.lastTags

    def factsFromCards(self, cards):
        "Find facts in a list of cards."
        (singles, pairs, fields) = self.getGroupedCards(cards)
        facts = []
        seen = {}
        for card in pairs:
            if seen.get(id(card)):
                continue
            for field in self.model.unique:
                try:
                    idx = self.mapping.index(field)
                    data = card[idx].lower()
                except ValueError:
                    # field is not mapped, ignore
                    continue
                cards = fields[field][data]
                if len(cards) > 1:
                    facts.append(ForeignFact(*cards))
                    for card in cards:
                        seen[id(card)] = True
                    break
        # add all the singles at the end
        facts.extend([ForeignFact(card) for card in singles])
        return facts

    def getGroupedCards(self, cards):
        "Override this to force non-grouping."
        if len(self.activatedCards) < 2:
            # only a single card model is available, and we can't do
            # lexical analysis. don't group
            (singles, pairs, fields) = self.groupCards(cards)
            return (singles + pairs, [], fields)
        if not self.autoSingle:
            return self.groupCards(cards)
        else:
            (singles, pairs, fields) = self.groupCards(cards)
            ratio = (len(singles) + 1) / float(len(pairs) + 1)
            if ratio < 1.7:
                return (singles, pairs, fields)
            else:
                return (singles + pairs, [], fields)

    def groupCards(self, cards):
        "Sort CARDS into a list of single entries + reverse-entry pairs."
        fields = {}
        ingroup = {}
        for field in self.model.unique:
            fields[field] = {}
        for card in cards:
            # for each field of the card
            for num in range(len(card)):
                # if it's also unique in the model
                for field in self.model.unique:
                    # check to see if we've already seen it
                    match = fields[field].get(card[num].lower())
                    if match:
                        # if we've seen it once, mark as a group
                        if len(match) == 1:
                            match.append(card)
                            for c in match:
                                ingroup[id(c)] = True
                    else:
                        # mark it as seen for later
                        if self.mapping[num] == field:
                            fields[field][card[num].lower()] = [card]
        singles = [card for card in cards if id(card) not in ingroup]
        pairs = [card for card in cards if id(card) in ingroup]
        return (singles, pairs, fields)

    def adjustModelCards(self, ffact):
        "Select the appropriate cards for foreign fact FFACT."
        self.model.cards = self.activatedCards[:]
        if len(self.model.cards) < 2:
            # don't try and remove single cards when there's only one available
            self.singles += 1
            return
        if len(ffact) == 1 and self.autoSingle:
            # the card is a single, and we're trying to preserve the old
            # structure
            self.singles += 1
            try:
                if (ffact[0].matchingTag == "Recognition" or not
                    ffact[0].matchingTag):
                    self.model.disableCard("Production")
                elif ffact[0].matchingTag == "Production":
                    self.model.disableCard("Recognition")
            except IndexError:
                print "index error"
                # couldn't preserve, as it wasn't a standard ja/en deck
                pass
        else:
            self.multiples += len(ffact)

    def addFact(self, ffact):
        self.adjustModelCards(ffact)
        # make sure we derive the fact after the model has been set
        fact = self.factFromForeign(ffact)
        model = self.deck.facts.models.current
        req = model.required
        uniq = model.unique
        # even if the model states that a field is required, the user probably
        # wants to import anyway, so we check and tag
        try:
            model.ensureValid(fact)
        except FactInvalidError, e:
            fact.tags.append(_("Import: missing field"))
            self.missing += 1
        model.required = []
        # likewise with duplicates
        try:
            model.ensureUnique(fact, self.deck.facts)
        except FactInvalidError, e:
            fact.tags.append(_("Import: duplicate"))
            fact.tags.append(_("Import: same %s") % e.data['field'])
            self.dupes += 1
        model.unique = []
        # do import
        cards = self.deck.addFact(fact)
        self.addedFacts.append(fact)
        model.required = req
        model.unique = uniq
        self.copyMeta(cards, ffact)

    def factFromForeign(self, ffact):
        "Return a fact given a foreign fact FFACT."
        fact = Fact()
        ffields = ffact.fields()
        if len(self.mapping) != len(ffields):
            raise ImportFormatError(type="mappingSizeWrong",
                                    info="got %d fields, expected %d" %
                                    (len(self.mapping), len(ffields)))
        # map the fields
        for n in range(len(self.mapping)):
            field = self.mapping[n]
            if not field:
                continue
            fact[field] = ffields[n]
        # set any remaining fields to blank
        for field in self.deck.facts.models.current.fields:
            if not field.name in fact:
                fact[field.name] = ""
        return fact

    def copyMeta(self, cards, ffact):
        "Copy metadata for each card based on FFACT."
        mapping = []
        if len(cards) == 1:
            mapping.append((ffact[0], cards[0]))
        else:
            for fcard in ffact:
                found = False
                for card in cards:
                    # for japanese
                    if fcard.matchingTag == card.model.name:
                        mapping.append((fcard, card))
                        found = True
                        break
                    # for standard cards
                    if (fcard[0] == card.question or
                        fcard[1] == card.answer):
                        mapping.append((fcard, card))
                        found = True
                        break
                if not found:
                    # we lost some progress data due to a duplicate
                    pass
        # copy the metadata from the foreign cards into our generated ones
        for (fcard, card) in mapping:
            self.copyCardMeta(fcard, card)

    def copyCardMeta(self, fcard, card):
        "Copy generic and scheduling metadata."
        for attr in ('created', 'modified', 'tags', 'stats'):
            if hasattr(fcard, attr):
                # avoid changing 'modified' with setattr
                card.__dict__[attr] = getattr(fcard, attr)
        self.deck.sched.copyToCard(fcard, card)

# Base importer
##########################################################################

class Importer(object):

    needModel = True

    def __init__(self, file, model, mapping=None):
        self.file = file
        self._model = model
        self._mapping = mapping
        self.log = []

    def getImporter(self, deck, model, mapping):
        return ForeignCardImporter(deck, model, mapping)

    def importInto(self, deck):
        "Import into DECK. Return an importer object with a log."
        cards = self.foreignCards()
        importer = self.getImporter(deck, self.model, self.mapping)
        importer.importCards(cards)
        log = self.log
        log.extend(importer.log)
        importer.log = log
        self.importer = importer
        return importer

    def foreignCards(self):
        "Return a list of foreign cards for importing."
        return []

    def resetMapping(self):
        "Reset mapping to default and return it."
        numFields = self.fields()
        mapping = self.model.mapping()
        map = [None] * numFields
        for n in range(len(mapping)):
            if n == numFields:
                break
            map[n] = mapping[n]
        self._mapping = map
        return map

    def getMapping(self):
        if not self._mapping:
            self.resetMapping()
        return self._mapping

    def setMapping(self, mapping):
        self._mapping = mapping

    mapping = property(getMapping, setMapping)

    def getModel(self):
        return self._model

    def setModel(self, model):
        self._model = model
        # update the mapping for the new model
        self._mapping = None
        self.getMapping()

    model = property(getModel, setModel)

    def fields(self):
        "The number of fields."
        assert not "Not implemented"

# Japanese support for old Anki decks & pending files
##########################################################################

# format:
# kanji : (hiragana) english
# english : (hiragana) kanji

def isJapaneseText(text):
    "True if 70% of text is a Japanese character."
    total = len(text)
    if total == 0:
        return True
    jp = 0
    en = 0
    for c in text:
        if ord(c) >= 0x2E00 and ord(c) <= 0x9FFF:
            jp += 1
        if re.match("[A-Za-z]", c):
            en += 1
    if not jp:
        return False
    return ((jp + 1) / float(en + 1)) >= 1.0

def addJapaneseFields(fcard, qanda):
    "Try to guess kanji/hiragana from old-style Anki format."
    fields = {}
    tag = None
    ans = qanda[1]
    m = re.search("^(.+) \((.*)\)$", ans)
    if m and isJapaneseText(m.group(2)) and not isJapaneseText(qanda[0]):
        # kanji (hiragana)
        tag = "Production"
        fields['kanji'] = m.group(1)
        fields['hiragana'] = m.group(2)
        fields['misc'] = qanda[0]
    else:
        m = re.search("^\((.*?)\) (.+)$", ans)
        if m and isJapaneseText(m.group(1)):
            # misc (hiragana)
            tag = "Recognition"
            fields['hiragana'] = m.group(1)
            fields['misc'] = m.group(2)
            fields['kanji'] = qanda[0]
    if not fields:
        if isJapaneseText(qanda[0]):
            fields['kanji'] = qanda[0]
            fields['misc'] = qanda[1]
            fields['hiragana'] = ""
            tag = "Recognition"
        else:
            fields['kanji'] = qanda[1]
            fields['misc'] = qanda[0]
            fields['hiragana'] = ""
            tag = "Production"
    # user may have hiragana in both sections in old decks
    if fields['kanji'] == fields['hiragana']:
        fields['hiragana'] = ""
    fcard.append(fields['kanji'])
    fcard.append(fields['misc'])
    fcard.append(fields['hiragana'])
    fcard.matchingTag = tag

# Old Anki decks
##########################################################################

class AnkiV2Importer(Importer):

    def __init__(self, *args):
        Importer.__init__(self, *args)
        # anki 0.2.x only supported these two languages
        if "Japanese" in self.model.tags:
            self.format = "jp"
        else:
            self.format = "en"

    def getImporter(self, deck, model, mapping):
        imp = ForeignCardImporter(deck, model, mapping)
        # enable single detection
        imp.autoSingle = True
        return imp

    def foreignCards(self):
        # load pickle file, using dummy classes
        ds = anki.Controller.DeckStorage()
        oldDeck = ds.fromFile(self.file, dummyClasses=True)
        self.oldDeck = oldDeck
        # generate foreign cards
        fcards = []
        for card in oldDeck:
            fcard = ForeignCard()
            fcard.__dict__.update(card.__dict__)
            self.updateCard(fcard)
            if self.format == "en":
                self.addFields(fcard, (card.question, card.answer))
            else:
                addJapaneseFields(fcard, (card.question, card.answer))
            # sanatize the data, which may be wrong in previous versions
            fcard.firstAnswered = float(getattr(fcard, 'firstAnswered', 0) or 0)
            fcard.lastInterval = float(
                getattr(fcard, 'lastInterval', fcard.interval))
            # and add it to our list
            fcards.append(fcard)
        return fcards

    def addFields(self, fcard, (q, a)):
        "Add fields, trying to place English in first entry."
        # one sided decks should honour order
        if len(self.model.cards) < 2:
            fcard.append(q)
            fcard.append(a)
        elif isJapaneseText(q) and not isJapaneseText(a):
            fcard.append(a)
            fcard.append(q)
            fcard.matchingTag = "Recognition"
        elif isJapaneseText(a) and not isJapaneseText(q):
            fcard.append(q)
            fcard.append(a)
            fcard.matchingTag = "Production"
        else:
            fcard.append(q)
            fcard.append(a)

    def fields(self):
        if self.format == "jp":
            return 3
        else:
            return 2

    def copyStats(self, deck):
        "Copy the stats into DECK. Any existing stats will be overwritten."
        # this assumes the 'sched' scheduler. we don't call it by default.
        if hasattr(self.oldDeck.manager, "easeStats"):
            deck.sched.easeStats = self.oldDeck.manager.easeStats
        if hasattr(self.oldDeck.manager, "globalStats"):
            for type in deck.sched.cardTypes:
                deck.sched.globalStats[type] = (
                    self.oldDeck.manager.globalStats[type])

    def updateCard(self, card):
        "Change to 0.3 card format."
        card.stats['young']['yes'] = card.correctCount
        card.stats['young']['no'] = card.totalCount - card.correctCount
        card.stats['successivelyCorrect'] = card.successivelyCorrect

def upgradeV2(deckPath, newModel):
    dir = os.path.dirname(deckPath)
    file = os.path.basename(deckPath)
    (file, ext) = os.path.splitext(file)
    ext = ext.lower().replace(".fc", "")
    ext += ".anki"
    while True:
        newPath = os.path.join(dir, file+ext)
        if not os.path.exists(newPath):
            break
        file += "1"
    # load the old file
    deck = Controller.Deck()
    del deck.facts.models[0]
    imp = anki.importing.AnkiV2Importer(deckPath, newModel)
    imp.importInto(deck)
    if imp.importer.singles and not imp.importer.multiples:
        # automatically switched to singles mode. reimport again without
        # reordering
        deck = Controller.Deck()
        del newModel.cards[0]
        del deck.facts.models[0]
        newModel.unique.remove("Meaning")
        imp.importInto(deck)
    imp.copyStats(deck)
    # write to a new name
    Controller.DeckStorage().save(deck, newPath)
    return deck

# Mnemosyne
##########################################################################

class MnemosyneImporter(Importer):

    def foreignCards(self):
        # empty objects so we can load the native mnemosyne file
        class MnemosyneModule(object):
            class StartTime:
                pass
            class Category:
                pass
            class Item:
                pass
        for module in ('mnemosyne',
                       'mnemosyne.core',
                       'mnemosyne.core.mnemosyne_core'):
            sys.modules[module] = MnemosyneModule()
        try:
            file = open(self.file, "rb")
        except (IOError, OSError), e:
            raise ImportFormatError(type="systemError",
                                    info=str(e))
        file.readline()
        # read the structure in
        try:
            struct = pickle.load(file)
        except (EOFError, KeyError):
            raise ImportFormatError(type="invalidFile")
        startTime = struct[0].time
        daysPassed = (time.time() - startTime) / 86400.0
        # gather cards
        cards = []
        for item in struct[2]:
            card = ForeignCard()
            card.stats = Controller.Deck().sched.newStats()
            # add text
            card.append(self.fudgeText(item.q))
            card.append(self.fudgeText(item.a))
            # scheduling data
            card.interval = item.next_rep - item.last_rep
            secDelta = (item.next_rep - daysPassed) * 86400.0
            card.nextTime = time.time() + secDelta
            # for some reason mnemosyne starts cards off on 1 instead of 0
            card.stats['successivelyCorrect'] = max(
                (item.acq_reps_since_lapse + item.ret_reps_since_lapse -1), 0)
            card.stats['young']['yes'] = (item.acq_reps + item.ret_reps) - 1
            card.stats['young']['no'] = item.lapses
            if item.cat.name != u"<default>":
                card.tags.append(item.cat.name)
            cards.append(card)
        return cards

    def fields(self):
        return 2

    def fudgeText(self, text):
        text = text.replace("\n", "<br>")
        text = re.sub('<sound src="(.*?)">', '[sound:\\1]', text)
        return text

# Text files
##########################################################################

class TextImporter(Importer):

    patterns = (":", "\t", ";")

    def __init__(self, *args):
        Importer.__init__(self, *args)
        self.lines = None

    def readFile(self):
        f = codecs.open(self.file, encoding="utf-8")
        try:
            data = f.readlines()
        except UnicodeDecodeError, e:
            raise ImportFormatError(type="encodingError",
                                    info=_("The file was not in UTF8 format."))
        if not data:
            return []
        if data[0].startswith(unicode(codecs.BOM_UTF8, "utf8")):
            data[0] = data[0][1:]
        # remove comment char
        lines = [l for l in data if not l.lstrip().startswith("#")]
        return lines

    def fields(self):
        "Number of fields."
        self.parseTopLine()
        return self.numFields

    def cacheFile(self):
        "Read file into self.lines if not already there."
        if not self.lines:
            self.lines = self.readFile()

    def parseTopLine(self):
        "Parse the top line and determine the pattern and number of fields."
        # load & look for the right pattern
        self.cacheFile()
        # actually, we look for the first non-blank line
        lines = [l for l in self.lines if l.strip()]
        if not lines:
            raise ImportFormatError(type="emptyFile",
                                    info=_("The file had no non-empty lines."))
        line = lines[0]
        found = False
        for p in self.patterns:
            if p in line:
                pattern = p
                fields = line.split(p)
                numFields = len(fields)
                found = True
                break
        if not found:
            fmtError = _(
                "Couldn't find pattern. The file should be a series "
                "of lines separated by colons, semicolons or tabs.")
            raise ImportFormatError(type="invalidPattern",
                                    info=fmtError)
        self.pattern = pattern
        self.setNumFields(line)

    def setNumFields(self, line):
        self.numFields = len(self.parseLine(line))

    def parseLine(self, line):
        fields = line.split(self.pattern)
        fields = [f.strip() for f in fields]
        return fields

    def foreignCards(self):
        self.parseTopLine()
        # process all lines
        log = []
        cards = []
        lineNum = 0
        ignored = 0
        for line in self.lines:
            lineNum += 1
            if not line.strip():
                # ignore blank lines
                continue
            try:
                fields = self.parseLine(line)
            except ValueError:
                log.append(_("Line %(line)d doesn't match pattern '%(pat)s'")
                           % {
                    'line': lineNum,
                    'pat': pattern,
                    })
                ignored += 1
                continue
            if len(fields) != self.numFields:
                log.append(_(
                    "Line %(line)d had %(num1)d fields,"
                    " expected %(num2)d") % {
                    "line": lineNum,
                    "num1": len(fields),
                    "num2": self.numFields,
                    })
                ignored += 1
                continue
            card = self.cardFromFields(fields)
            cards.append(card)
        self.log = log
        self.ignored = ignored
        return cards

    def cardFromFields(self, fields):
        card = ForeignCard()
        card.extend(fields)
        return card

# Old Fc .pending files
##########################################################################

class FcPendingImporter(TextImporter):

    def cardFromFields(self, fields):
        card = ForeignCard()
        addJapaneseFields(card, fields)
        return card

    def setNumFields(self, line):
        self.numFields = 2

    def fields(self):
        return 3

    def getImporter(self, deck, model, mapping):
        imp = ForeignCardImporter(deck, model, mapping)
        # enable single detection
        imp.autoSingle = True
        return imp

# Anki decks
##########################################################################

# this is a bit hackish, due to the inflexible importing design
class AnkiImporter(Importer):

    needModel = False

    def __init__(self, *args):
        Importer.__init__(self, *args)

    def importInto(self, deck):
        "Import into DECK. Return an importer object with a log."
        self.deck = deck
        ds = Controller.DeckStorage()
        self.importer = DummyCardImporter()
        ideck = ds.load(self.file)
        for card in ideck:
            if card not in deck:
                self.addModel(card.fact.model)
                self.addFact(card.fact)
                self.addCard(card)
        self.importer.total = len(self.importer.addedFacts)
        return self.importer

    def fields(self):
        return 0

    def addModel(self, model):
        if model not in self.deck.facts.models:
            self.deck.facts.models.addModel(model)

    def addFact(self, fact):
        if fact in self.deck.facts:
            return
        model = fact.model
        req = model.required
        uniq = model.unique
        try:
            model.ensureValid(fact)
        except FactInvalidError, e:
            fact.tags.append(_("Import: missing field"))
            self.importer.missing += 1
        model.required = []
        try:
            model.ensureUnique(fact, self.deck.facts)
        except FactInvalidError:
            fact.tags.append(_("Import: duplicate"))
            self.importer.dupes += 1
        model.unique = []
        # do import
        self.deck.facts.addFact(fact, validate=False)
        self.importer.addedFacts.append(fact)
        model.required = req
        model.unique = uniq

    def addCard(self, card):
        self.deck.addCard(card, len(self.deck))

# emulate the card importing information
class DummyCardImporter(object):
    def __init__(self):
        self.dupes = 0
        self.missing = 0
        self.addedFacts = []
        self.total = 0
        self.log = []

# Export modules
##########################################################################

Importers = (
    (_("Colon/semicolon/tab-separated text file (*.*)"), TextImporter),
    (_("Anki deck (*.anki)"), AnkiImporter),
    (_("Old Anki 0.2.x file (*.fc)"), AnkiV2Importer),
    (_("Fc .pending file (*.pending *.imported)"), FcPendingImporter),
    (_("Mnemosyne file (*.mem)"), MnemosyneImporter),
    )
