#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

"""\
Model - define the way in which facts are added and shown
==============================================================
"""

import time, sys, re
from anki.lang import _
from anki.cards import Card
from anki.decorators import Decorators
from anki.errors import *
from anki.fields import *
from anki.cardmodels import *
from anki.ids import *

# Due to backward compatibility, decks may have n > 1 for a while
Positions = {
    0: _("Random"),
    1: _("Order by input date"),
    }

Alignment = {
    0: _("Center"),
    1: _("Left"),
    2: _("Right"),
    }

# Models
##########################################################################

class Models(IDList):
    "An order-independent list of models."

    def __init__(self, parent=None):
        IDList.__init__(self, parent)
        self.current = None

    def setModel(self, model):
        "Select MODEL as current, adding to models if necessary."
        found = False
        for m in self:
            # times may be the same, so compare live object ids
            if id(m) == id(model):
                found = True
                break
        if not found:
            self.addModel(model)
        self.current = model
        self.setModified()

    def addModel(self, model):
        self.append(model)
        self.makeModelUnique(model)

    def makeModelUnique(self, model):
        "Give model a unique model name based on the existing name."
        origName = model.name
        n = 1
        while 1:
            if self.modelNameIsUnique(model.name):
                break
            else:
                model.name = "%s<%d>" % (origName, n)
                n += 1

    def byName(self, modelName):
        for model in self:
            if model.name == modelName:
                return model
        return None

    def modelNameIsUnique(self, modelName):
        "Called after model is inserted."
        return len([m for m in self if m.name == modelName]) <= 1

    def __unicode__(self):
        s = u"<models models=%d curmodel='%s'>" % (
            len(self), self.current)
        return s

# Information model: constrains facts and define cards & layout
##########################################################################

class Model(IDObject):
    "Defines the way a fact behaves, what fields it can contain, etc."

    def __init__(self, name="", description=""):
        IDObject.__init__(self)
        if not name:
            name = _("Model")
        self.name = name
        self.description = description
        self.required = []
        self.unique = []
        self.fields = Fields(self)
        self.allcards = CardModels(self)
        self.cards = []
        self.tags = []
        self.decorators = []
        self.position = 0

    def __unicode__(self):
        return "<model '%s'>" % self.name

    # Fields
    ##########################################################################

    def addField(self, field):
        self.fields.append(field)

    def ensureValid(self, fact):
        "Raise an error if missing fields are found in FACT."
        for field in self.fields:
            if field.name not in fact:
                raise FactInvalidError(type="fieldNotDefined",
                                       field=field.name)
        for fieldName in self.required:
            if not self.fieldIsValid(fact, fieldName):
                raise FactInvalidError(type="fieldEmpty",
                                       field=fieldName)

    def ensureUnique(self, fact, facts):
        "Raise an error if duplicate fields are found in FACT."
        for fieldName in self.unique:
            if not self.fieldIsUnique(fact, fieldName, facts):
                raise FactInvalidError(type="fieldNotUnique",
                                       field=fieldName)

    def fieldIsValid(self, fact, fieldName):
        "True if FIELDNAME of FACT is non-empty if required."
        return fact.get(fieldName, "").strip()

    def fieldIsUnique(self, fact, fieldName, facts):
        "True if FIELDNAME of FACT is unique in FACTS. Ignore if empty."
        val = fact.get(fieldName, "").strip()
        if not val:
            return True
        found = [f for f in facts.byField(fieldName, val)
                 if f.id and f.id != fact.id]
        return not found

    def fieldByName(self, name):
        "Return field with NAME. Return None if not found."
        fields = [f for f in self.fields if f.name == name]
        if len(fields) == 1:
            return fields[0]
        return None

    def mapping(self):
        "Return a default mapping for use in importing."
        # add required fields first
        fields = [f for f in self.fields if f.name in self.required]
        # and any remaining fields
        for field in self.fields:
            if field not in fields:
                fields.append(field)
        return [f.name for f in fields]

    # Cards
    ##########################################################################

    def makeCards(self, fact):
        "Return a list of active cards for FACT."
        return [cmodel.makeCard(fact) for cmodel in self.cards]

    def activateAllCards(self):
        "Make all cards active."
        self.cards = self.allcards[:]
        self.setModified()

    def disableCard(self, name):
        "Disable card with NAME."
        card = [card for card in self.cards if card.name == name][0]
        self.cards.remove(card)
        self.setModified()

    def cardsByName(self, name):
        return [c for c in self.allcards if c.name == name]

    def removeCardModel(self, cardModel):
        self.allcards.remove(cardModel)
        if cardModel in self.cards:
            self.cards.remove(cardModel)
        self.setModified()

    def addCard(self, card):
        "Add card model CARD."
        self.allcards.append(card)
        self.cards.append(card)

    # Decorators
    ##########################################################################

    def onKeypress(self, fact, field, value):
        "Do per-keypress processing."
        self.withDecorators("onKeypress", fact, field, value)

    def onSubmit(self, fact):
        "Process fact before it is added."
        # strip leading/trailing whitespace and add any missing fields
        for field in self.fields:
            fact[field.name] = fact.get(field.name, "").strip()
        self.withDecorators("onSubmit", fact)

    def withDecorators(self, cmd, *args):
        "Run CMD on all decorators that are available and support it."
        Decorators.doAll(cmd, self.decorators, *args)
