#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

"""\
The standard Anki scheduler
==============================
"""
__docformat__ = 'restructuredtext'

import time
import copy
import random
import math

from anki.lang import _, ngettext
from anki.errors import *

import anki.utils, anki.deck, anki.lang, anki.cards

class Scheduler(object):
    "A strategy for scehduling cards using the Anki algorithm."

    def __init__(self):
        self.deck = None
        self.schedVersion = 0
        # if the card is "back to the start"
        self.startInterval = 0
        self.initialFactor = 1.5
        self.initalFactorChange = 0.05
        # the various scheduling choices when a card is at the start
        self.hardInterval = (1.0, 1.0)
        self.midInterval = (3.0, 5.0)
        self.easyInterval = (7.0, 9.0)
        # an additional multipler applied to cards marked easy
        self.factorFour = 1.5
        # default to a max scheduling of 5 years
        self.maxScheduleTime = 1825
        # delays on failure
        self.delay0 = 600
        self.delay1 = 600
        self.delay2 = 28800
        # collapsing
        self.collapse = 18000
        # priorities & postponing
        self.highPriority = []
        self.medPriority = []
        self.lowPriority = []
        self.suspendedTags = ["Suspended"]
        # define the card types for statistics
        self.cardTypes = ("new", "young", "old")
        # stats
        self.globalStats = self.newStats()
        self.easeStats = {}
        for i in self.cardTypes:
            self.easeStats[i] = {}
        self.dailyStats = {}
        # per load
        self.afterLoad()

    # Backend - factor/interval management
    ##########################################################################

    def nextInterval(self, card, ease):
        "Return the next interval for CARD given EASE."
        factor = 0
        delay = max(0, (time.time() - card.nextTime) / 86400.0)
        dev = getattr(card, 'deviation', 0)
        if self.cardIsInInitialState(card):
            if ease <= 2:
                interval = self.getYoungInterval(self.hardInterval)
            elif ease == 3:
                interval = self.getYoungInterval(self.midInterval)
            elif ease == 4:
                interval = self.getYoungInterval(self.easyInterval)
            else:
                raise ValueError("Invalid ease")
        elif ease == 0:
            interval = self.startInterval
        else:
            if ease == 4:
                factor = card.factor * self.factorFour + dev
            elif ease == 3:
                factor = (1 + card.factor * self.factorFour) / 2.0 + dev
            elif ease == 2:
                factor = 1 / card.factor
            elif ease == 1:
                factor = 1 / card.factor / 2.0
            else:
                raise ValueError("Invalid ease")
            interval = (card.interval + delay) * factor
        if not self.cardIsNew(card):
            # cards in the initial state need more conservative boosts
            if self.cardIsInInitialState(card) and delay > 3:
                delay = max(3, delay / 2.0)
            if ease == 4:
                interval += delay
            if ease == 3:
                interval += delay / 2.0
            if ease == 2 and not self.cardIsInInitialState(card):
                interval += delay / 2.0
        if self.maxScheduleTime:
            interval = min(self.maxScheduleTime, interval)
        #print "e=%d li=%0.2f ci=%0.2f, ni=%0.2f f=%0.2f d=%0.2f" % (
        #    ease, card.lastInterval, card.interval, interval, factor, delay)
        return interval

    def nextFactor(self, card, ease):
        "Return the next factor for CARD given EASE."
        if ease == 4:
            factor = card.factor + (card.factorChange * 2)
        elif ease == 3:
            factor = card.factor + card.factorChange
        elif ease == 2:
            factor = card.factor - card.factorChange
        elif ease == 1:
            factor = card.factor - (card.factorChange * 2)
        elif ease == 0:
            factor = self.initialFactor
        else:
            raise ValueError("Invalid ease")
        factor = max(1.05, factor)
        factor = min(2.00, factor)
        return factor

    def scheduleCard(self, card, ease):
        "Update factor, interval and reschedule."
        wasInitial = self.cardIsInInitialState(card)
        card.lastInterval = card.interval
        card.interval = self.nextInterval(card, ease)
        card.factor = self.nextFactor(card, ease)
        card.lastTime = card.nextTime
        if ease < 2:
            # mistake, redisplay
            card.nextTime = time.time()
            if ease == 0:
                delay = self.delay0
            else:
                # use separate delay for mature cards
                if wasInitial:
                    delay = self.delay1
                else:
                    delay = self.delay2
            card.nextTime += delay
            card.lastInterval = delay / 86400.0
            newInterval = 0
        else:
            card.nextTime = time.time() + card.interval * 86400
            newInterval = card.interval
        if self.cardIsInInitialState(card):
            card.factor = self.initialFactor
        return newInterval

    def deviatedCard(self, card):
        "Return a card with a +/- deviation to ensure even spacing."
        a = 0.047; b = 0.092; p = random.random() - 0.5
        m = -1.0 / b * math.log(1.0 - b / a * abs(p))
        # randomly choose + or -
        if p < 0:
            m *= -1
        else:
            m *= 1
        m = ((100+m) / 100.0) - 1
        # set the deviation, don't update modified time
        card.__dict__['deviation'] = m
        return card

    def getYoungInterval(self, minMax):
        "Return an interval in the range minMax."
        return random.uniform(*minMax)

    # Retrieving pending cards
    ##########################################################################

    def getCard(self):
        "Get the next waiting old card in the deck, an unseen one, or None."
        # get all pending cards
        cards = self.getCards()
        if not cards:
            return None
        self.lastCardRetrievedTime = time.time()
        return self.deviatedCard(cards[0])

    def getCards(self):
        "Return a sorted list of all pending cards."
        cards = [c for c in self.deck if self.cardIsPending(c)]
        cards.sort(cmp=self.cmpStandardOrder)
        if not cards:
            # try again with pending cards
            cards = [c for c in self.deck if self.cardIsPending(c, True)]
            # and order by pure delay
            cards.sort(cmp=self.cmpPendingOrder)
        return cards

    def cmpStandardOrder(self, a, b):
        # separate high/low priority
        ret = cmp(b.priority == anki.cards.PRIORITY_HIGH,
                  a.priority == anki.cards.PRIORITY_HIGH)
        if ret != 0:
            return ret
        ret = cmp(a.priority == anki.cards.PRIORITY_LOW,
                  b.priority == anki.cards.PRIORITY_LOW)
        if ret != 0:
            return ret
        # separate old/new
        ret = cmp(self.cardIsNew(a), self.cardIsNew(b))
        if ret != 0:
            return ret
        # separate failed/expired cards
        ret = cmp(self.cardWasFailed(b), self.cardWasFailed(a))
        if ret != 0:
            return ret
        # separate medium/non-medium priority
        ret = cmp(b.priority == anki.cards.PRIORITY_MED,
                  a.priority == anki.cards.PRIORITY_MED)
        if ret != 0:
            return ret
        # if the cards are new, use deck order
        if self.cardIsNew(a):
            return 0
        # else order by relative delay
        return cmp(self.relativeCardDelay(a), self.relativeCardDelay(b))

    def cmpPriority(self, list, a, b, transform=1):
        aMatch = False
        bMatch = False
        for tag in list:
            if a.hasTag(tag):
                aMatch = True
            if b.hasTag(tag):
                bMatch = True
            if aMatch and bMatch:
                # short circuit
                break
        return cmp(bMatch, aMatch) * transform

    def cmpPendingOrder(self, a, b):
        "Order by mod time for end of session work."
        return cmp(a.modified, b.modified)

    # Cards - adding/answering
    ##########################################################################

    def addToCard(self, card, adjustPending=True):
        """Add scheduling info to CARD."""
        # only update information if it doesn't already exist
        if not getattr(card, "factor", None):
            card.nextTime = time.time()
            card.lastTime = 0
            card.interval = self.startInterval
            card.lastInterval = self.startInterval
            card.factor = self.initialFactor
            card.factorChange = self.initalFactorChange
            card.firstAnswered = 0
            card.stats = self.newStats()
            card.stats['successivelyCorrect'] = 0
            card.total = 0
        if adjustPending and card.nextTime < time.time():
            self.adjustPending(card)

    def copyToCard(self, card1, card2):
        "Copy scheduling info from CARD1 to CARD2."
        for attr in ("nextTime", "interval", "lastInterval",
                     "factor", "factorChange", "firstAnswered",
                     "stats", "lastTime", "total"):
            card2.__dict__[attr] = getattr(card1, attr)

    def resetCard(self, card):
        self.updateHistory(card, type="reset")
        card.factor = None
        self.addToCard(card, adjustPending=False)

    def makeDue(self, card):
        self.updateHistory(card, type="makeDue")
        card.interval = card.lastInterval
        card.nextTime = card.lastTime

    def answerCard(self, card, ease):
        "Reschedule CARD based on EASE."
        self.adjustPending(card, -1)
        # get the card type before scheduling
        type = self.cardType(card)
        next = self.scheduleCard(card, ease)
        self.updateStats(card, ease, type)
        self.updateHistory(card, ease)
        card.fact.cardAnswered(card)
        return next

    def updateHistory(self, card, ease=None, type="answered"):
        d = {
            "lastInterval": card.lastInterval,
            "interval": card.interval,
            "ease": ease,
            "time": time.time(),
            }
        if card.lastTime:
            d["delay"] = time.time() - card.lastTime
        else:
            d["delay"] = 0
        card.history.append((type, d))

    def cardIsNew(self, card):
        "True if a card has never been seen before."
        # hack for old clients syncing against us
        return getattr(card, 'total', 0) == 0

    def cardIsInInitialState(self, card):
        "True if card hasn't been scheduled for easyInterval yet."
        return card.interval < self.easyInterval[0]

    def cardIsYoung(self, card):
        "True if card hasn't been remembered for at least easyInterval."
        if card.interval == card.lastInterval:
            # upgraded deck doesn't have proper interval yet - so to determine
            # if a card has been correctly answered after easyInterval, we
            # assume the next interval has been scheduled for at least 12
            # days. this is not accurate, which is why we want to move
            # to using lastInterval
            return card.interval < 12
        else:
            # card's current interval is out of the initial state, and the
            # last interval was over easyInterval
            return self.cardIsInInitialState(card) or \
                   card.lastInterval < self.easyInterval[0]

    def relativeCardDelay(self, card):
        "Return the delay of a card relative to a card's interval."
        return card.interval / ((time.time() - card.nextTime) / 86400.0)

    def cardWasFailed(self, card):
        "True if card was last answered incorrectly."
        return (card.lastTime != 0 and
                not card.stats['successivelyCorrect'])

    # Pending cards - seen/unseen
    ##########################################################################

    def getPending(self):
        if not self.pendingOld:
            self.resetPending()
        if (self.pendingOld % 10) == 0:
            self.resetPending()
        return (self.pendingOld, self.pendingNew)

    def resetPending(self):
        "Update number of pending cards after deck edits."
        # FIXME: the need for this could be removed by checking deleted cards
        # and updated cards when syncing
        (self.pendingOld, self.pendingNew) = self.pendingCards()

    def pendingCards(self):
        """Return the number of pending cards, in a tuple (seen, unseen)."""
        if not self.deck:
            # called before we're attached to a deck
            return (0, 0)
        old=0
        new=0
        seenFacts = {}
        cards = [c for c in self.deck if self.cardIsPending(c)]
        cards2 = [c for c in self.deck if self.cardIsPending(c, True)]
        for card in cards:
            if card.fact.id in seenFacts:
                continue
            seenFacts[card.fact.id] = True
            if self.cardIsNew(card):
                new += 1
            else:
                old +=1
        if old or new:
            return (old, new)
        else:
            # nothing left. look ahead.
            return (len([c for c in self.deck
                         if self.cardIsPending(c, True)]), 0)


    def adjustPending(self, card, diff=1):
        "Update pendingNew or pendingOld for CARD."
        if self.cardIsNew(card):
            self.pendingNew += diff
        else:
            self.pendingOld += diff
        if (self.pendingOld < 0 or
            self.pendingNew < 0):
            self.resetPending()

    def cardIsPending(self, card, delayed=False):
        if delayed:
            offset = self.collapse
        else:
            offset = 0
        return (self.nextTime(card) <= (time.time() + offset)
                and not self.cardIsSuspended(card))

    def cardIsSuspended(self, card):
        return card.priority == anki.cards.PRIORITY_NONE

    def cardIsSpaced(self, card):
        return (card.nextTime <= time.time() and
                card.model.timeLeft(card))

    # Times
    ##########################################################################

    def earliestTime(self):
        """Return the time of the earliest card.
        This may be in the past if the deck is not finished.
        If the deck is empty, return None."""
        earliestCard = None
        for i in range(len(self.deck)):
            if self.cardIsSuspended(self.deck[i]):
                continue
            if not earliestCard:
                earliestCard = self.deck[i]
            else:
                if self.nextTime(self.deck[i]) < self.nextTime(earliestCard):
                    earliestCard = self.deck[i]
        if earliestCard:
            return self.nextTime(earliestCard)
        return

    def nextTime(self, card):
        left = card.model.timeLeft(card)
        next = card.nextTime
        if left:
            next = time.time() + left
        return max(card.nextTime, next)

    def earliestTimeStr(self, next=None):
        """Return the relative time to the earliest card as a string."""
        if next == None:
            next = self.earliestTime()
        if not next:
            return _("unknown")
        diff = next - time.time()
        return anki.utils.fmtTimeSpan(diff)

    def nextCardMsg(self):
        "Return absolute date and relative time to next card."
        import anki.lang
        relativeTime = self.earliestTime()
        relativeTimeStr = self.earliestTimeStr(relativeTime)
        local = time.strftime(_("%x at %H:%M"), time.localtime(relativeTime))
        return _("in %(a)s, on %(b)s") % {
            'a': relativeTimeStr,
            'b': local
            }

    def nextIntervalStr(self, card, ease):
        "Return the next interval for CARD given EASE as a string."
        if card.nextTime > time.time() and ease < 2:
            # the card is not yet due, and we are in the final drill
            return _("a short time")
        if ease == 0:
            interval = self.delay0 / 86400.0
        elif ease == 1:
            if self.cardIsInInitialState(card):
                interval = self.delay1 / 86400.0
            else:
                interval = self.delay2 / 86400.0
        elif self.cardIsInInitialState(card) and ease >= 2:
            if ease == 4:
                days = list(self.easyInterval)
            elif ease == 3:
                days = list(self.midInterval)
            else:
                days = list(self.hardInterval)
            if not self.cardIsNew(card):
                delay = (time.time() - card.nextTime) / 86400.0
                if delay > 3:
                    delay = max(3, delay / 2.0)
                if ease > 2:
                    # apply delay factor if it's not a new card
                    if ease == 3:
                        delay /= 2.0
                    days[0] += delay
                    days[1] += delay
            if days[0] != days[1]:
                return anki.utils.fmtTimeSpanPair(days[0]*86400.0,
                                                  days[1]*86400.0)
            else:
                interval = days[0]
        else:
            interval = self.nextInterval(card, ease)
        return anki.utils.fmtTimeSpan(interval*86400)

    def afterLoad(self):
        "Deck has been loaded from DeckStorage."
        self.maybeUpgrade()
        self.resetSessionStats()

    # Statistics - formatting and tracking
    ##########################################################################

    def resetSessionStats(self):
        "Reset correct/incorrect & pending cards at the start of a session."
        self.sessionStats = self.newStats()
        self.resetPending()

    def newStats(self):
        "Return a new stats object."
        stats = {
            'averageTime': 0,
            'totalTime': 0,
            }
        for type in self.cardTypes:
            stats[type] = {}
            for ans in ("yes", "no"):
                stats[type][ans] = 0
        return stats

    def answerCategory(self, ease):
        if ease < 2:
            return "no"
        else:
            return "yes"

    def cardType(self, card):
        if self.cardIsNew(card):
            return "new"
        elif self.cardIsYoung(card):
            return "young"
        else:
            return "old"

    def today(self):
        "The current day (0-indexed) relative to deck creation."
        # start the day at midnight
        initial = list(time.localtime(self.deck.created))
        initial[3:6] = (0,)*3
        initial = time.mktime(initial)
        now = time.time()
        return int((now - initial) / 86400.0)

    def daily(self):
        "Return the statistics for today, creating if necessary."
        t = self.today()
        if not self.dailyStats.has_key(t):
            self.dailyStats[t] = self.newStats()
            self.dailyStats[t]['pendingOnLoad'] = (
                self.pendingOld, self.pendingNew)
        return self.dailyStats[t]

    def allStatDicts(self, card=None):
        "A list of all stats (global, session, etc)"
        dicts = [self.sessionStats,
                 self.globalStats,
                 self.daily()]
        if card:
            dicts.append(card.stats)
        return dicts

    def updateStats(self, card, ease, type):
        "Update stats based on CARD and EASE."
        ans = self.answerCategory(ease)
        daily = self.daily()
        # mark correct/incorrect in the appropriate field
        for d in self.allStatDicts(card):
            d[type][ans] += 1
        # card specific
        if ans == "yes":
            card.stats['successivelyCorrect'] += 1
        else:
            card.stats['successivelyCorrect'] = 0
        if not card.firstAnswered:
            card.firstAnswered = time.time()
        card.total += 1
        # average ease answers
        self.easeStats[type][ease] = self.easeStats[type].get(ease, 0) + 1
        # average time taken to answer
        if getattr(self, 'lastCardRetrievedTime', None):
            timeTaken = time.time() - self.lastCardRetrievedTime
            # (we hope) long waits are not typical
            if timeTaken < 60:
                for d in self.allStatDicts(card):
                    if not d['totalTime']:
                        # fudge the numbers for imported decks
                        d['totalTime'] = timeTaken * self.getTotal(d)
                    else:
                        d['totalTime'] += timeTaken
                    d['averageTime'] = (d['totalTime'] /
                                        float(self.getTotal(d)))

    def summarizeStats(self, stats, pre=""):
        "Generate percentages and total counts for STATS. Optionally prefix."
        h = {}
        for j in self.cardTypes:
            for k in ("yes", "no"):
                # individual count
                h[pre + j.capitalize() + k.capitalize()] = stats[j][k]
                # type total
                h[pre + j.capitalize() + "Total"] = self.getTotalFromType(
                    stats, j)
        for j in ("yes", "no"):
            # yes/no totals
            num=0
            for k in self.cardTypes:
                num += h[pre + k.capitalize() + j.capitalize()]
            h[pre + j.capitalize() + "Total"] = num
        # total
        h[pre + "Total"] = self.getTotal(stats)
        for j in self.cardTypes:
            for k in ("yes", "no"):
                # individual percentages (eg gYoungYes / gYoungTotal)
                self._setPerc(h, pre + j.capitalize() + k.capitalize(),
                              pre + j.capitalize())
            # type perc (eg gOldYes / gOldTotal)
            self._setPerc(h, pre + j.capitalize() + "Total", pre + j.capitalize())
        for j in ("yes", "no"):
            # yes/no percs
            self._setPerc(h, pre + j.capitalize() + "Total", pre)
        h[pre+'TotalTime'] = anki.utils.fmtTimeSpan(stats['totalTime'],point=2)
        h[pre+'AverageTime'] = anki.utils.fmtTimeSpan(stats['averageTime'], point=2)
        return h

    def getStats(self, card=None):
        "Return a handy dictionary exposing a number of internal stats."
        # FIXME: fix 4am code
        h = {}
        h.update(self.summarizeStats(self.sessionStats, "s"))
        h.update(self.summarizeStats(self.globalStats, "g"))
        h.update(self.summarizeStats(self.daily(), "d"))
        if card:
            h.update(self.summarizeStats(card.stats, "c"))
        # pending cards
        p = self.getPending()
        h['pending'] = p[0] or p[1]
        h['learnMode'] = not p[0]
        (h['pendingOld'], h['pendingNew']) = p
        # estimated remaining time
        if self.sessionStats['averageTime']:
            h['timeLeft'] = anki.utils.fmtTimeSpan(
                self.sessionStats['averageTime'] * h['pending'],
                pad=0, point=1)
        else:
            h['timeLeft'] = _("Unknown")
        return h

    def getTotalFromType(self, stats, type):
        "Return total number of cards asked of TYPE."
        return stats[type]["yes"] + stats[type]["no"]

    def getTotal(self, stats):
        "Return total number of cards asked."
        total = 0
        for j in self.cardTypes:
            total += self.getTotalFromType(stats, j)
        return total

    def getTotalYes(self, stats, yes="yes"):
        total = 0
        for j in self.cardTypes:
            total += stats[j][yes]
        return total

    def getTotalNo(self, stats):
        return self.getTotalYes(stats, "no")

    def _setPerc(self, h, a, b):
        try:
            h[a + "%"] = (h[a] / float(h[b + "Total"])) * 100
        except ZeroDivisionError:
            h[a + "%"] = 0

    # Upgrading from previous decks
    ##########################################################################
    #
    # make sure to update bundleSched() in sync.py too!
    #

    def maybeUpgrade(self):
        # change old hardInterval from 1 day to 8-12 hours
        if list(self.hardInterval) == [1.0, 1.0]:
            self.hardInterval = [0.3333, 0.5]
        # add 'medium priority' support
        if not hasattr(self, 'medPriority'):
            self.medPriority = []
        # add delay2
        if not hasattr(self, "delay2"):
            self.delay2 = 28800
        # add collapsing
        if not hasattr(self, "collapse"):
            self.collapse = 18000
        # card related
        if self.deck is not None:
            # change priority handling
            self.deck.updatePriorities()
            # keep a tally of total reviews separately for efficiency.
            for card in self.deck:
                card.__dict__['total'] = self.getTotal(card.stats)
