#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

"""\
Standard models
==============================================================
"""

import sys
from anki.models import Model
from anki.fields import Field
from anki.cardmodels import CardModel
from anki.lang import _

models = {}

def byName(name):
    fn = models.get(name)
    if fn:
        return fn()
    raise ValueError("No such model available!")

def names():
    return models.keys()

# these are provided for convenience. all of the fields can be changed in real
# time and they will be stored with the deck.

# Other
##########################################################################

def OtherModel():
    m = Model(_('Other'),
              _('A basic flashcard with a front and a back.\n'
                'Questions are asked from front to back by default.\n\n'
                'Please consider customizing this model, rather than\n'
                'using it verbatim: field names like "expression" are\n'
                'clearer than "front" and "back", and will ensure\n'
                'that your entries are consistent.'))
    m.required = ['Front', 'Back']
    m.unique = ['Front', 'Back']
    m.addField(Field('Front', _('A question.')))
    m.addField(Field('Back', _('The answer.')))
    m.addCard(CardModel("Front to back", _("Front to back"),
                        "%(Front)s", "%(Back)s"))
    m.addCard(CardModel("Back to front", _("Back to front"),
                        "%(Back)s", "%(Front)s"))
    m.cards = [m.allcards[0]]
    return m
models['Other'] = OtherModel

# Japanese
##########################################################################

def JapaneseModel():
    m = Model(_("Japanese"),
              _("""
The reading field is automatically generated by default,
and shows the reading for the expression. For words that
are normally written in hiragana or katakana and don't
need a reading, you can put the word in the expression
field, and leave the reading field blank. A reading will
will not automatically be generated for words written
in only hiragana or katakana.

Note that the automatic generation of meaning is not
perfect, and should be checked before adding cards.""".strip()))
    m.required = ['Expression', 'Meaning']
    m.unique = ['Expression', 'Meaning']
    f = Field('Expression',
              _('A word or expression written in Kanji.'))
    font = "Mincho"
    f.display['quiz']['fontSize'] = 72
    f.display['quiz']['fontFamily'] = font
    f.display['edit']['fontFamily'] = font
    f.features.append("Reading source")
    m.addField(f)
    m.addField(Field('Meaning',
                          _(
        'A description in your native language, or Japanese')))
    f = Field('Reading',
              _('Automatically generated by default.'))
    f.display['quiz']['fontFamily'] = font
    f.display['edit']['fontFamily'] = font
    f.features.append("Reading destination")
    m.addField(f)
    m.addCard(CardModel("Production", _(
        "Actively test your recall by producing the target expression"),
                        "%(Meaning)s",
                        "%(Expression)s<br>%(Reading)s"))
    m.addCard(CardModel("Recognition", _(
        "Test your ability to recognize the target expression"),
                        "%(Expression)s",
                        "%(Reading)s<br>%(Meaning)s"))
    m.decorators.append("Japanese")
    m.tags.append("Japanese")
    return m
models['Japanese'] = JapaneseModel

# English
##########################################################################

def EnglishModel():
    m = Model(_("English"),
              _("""
Enter the English expression you want to learn in the 'Expression' field.
Enter a description in Japanese or English in the 'Meaning' field.""".strip()))
    m.required = ['Expression', 'Meaning']
    m.unique = ['Expression', 'Meaning']
    m.addField(Field('Expression'))
    m.addField(Field('Meaning'))
    m.addCard(CardModel(
        "Production", _("From the meaning to the English expression."),
        "%(Meaning)s", "%(Expression)s"))
    m.addCard(CardModel(
        "Recognition", _("From the English expression to the meaning."),
        "%(Expression)s", "%(Meaning)s"))
    # only use production by default
    del m.cards[1]
    m.tags.append("English")
    return m
models['English'] = EnglishModel

# Heisig
##########################################################################

def HeisigModel():
    m = Model(_("Heisig"),
              _("""
A format suitable for Heisig's "Remembering the Kanji".
You are tested from the keyword to the kanji.

Layout of the test is based on the great work at
http://kanji.koohii.com/

The link in the question will list user-contributed
stories. A free login is required.""".strip()))
    m.required = ['Kanji', 'Keyword']
    m.unique = ['Kanji', 'Keyword', 'Heisig number']
    m.position = 2
    font = "Mincho"
    f = Field('Kanji')
    f.display['quiz']['fontSize'] = 150
    f.display['quiz']['fontFamily'] = font
    f.display['edit']['fontFamily'] = font
    m.addField(f)
    m.addField(Field('Keyword'))
    m.addField(Field('Story'))
    m.addField(Field('Stroke count'))
    m.addField(Field('Heisig number'))
    m.addField(Field('Lesson number'))
    m.addCard(CardModel(
        "Production", _("From the keyword to the Kanji."),
        "<a href=\"http://kanji.koohii.com/study?framenum="
        "%(text:Heisig number)s\">%(Keyword)s</a><br>",
        "%(Kanji)s<br><table width=150><tr><td align=left>"
        u"画数%(Stroke count)s</td><td align=right>"
        "%(Heisig number)s</td></tr></table>"))
    m.tags.append("Heisig")
    return m
models['Heisig'] = HeisigModel

# Chinese: Mandarin & Cantonese
##########################################################################

def CantoneseModel():
    m = Model(_("Cantonese"),
              "")
    m.required = ['Expression', 'Meaning']
    m.unique = ['Expression', 'Meaning']
    f = Field('Expression',
              _('A word or expression written in Hanzi.'))
    f.display['quiz']['fontSize'] = 72
    f.features.append("Reading source")
    m.addField(f)
    m.addField(Field('Meaning',
                          _(
        'A description in your native language, or Cantonese')))
    f = Field('Reading',
              _('Automatically generated by default.'))
    f.features.append("Reading destination")
    m.addField(f)
    m.addCard(CardModel("Production", _(
        "Actively test your recall by producing the target expression"),
                        "%(Meaning)s",
                        "%(Expression)s<br>%(Reading)s"))
    m.addCard(CardModel("Recognition", _(
        "Test your ability to recognize the target expression"),
                        "%(Expression)s",
                        "%(Reading)s<br>%(Meaning)s"))
    m.decorators.append("Cantonese")
    m.tags.append("Cantonese")
    return m
models['Cantonese'] = CantoneseModel

def MandarinModel():
    m = Model(_("Mandarin"),
              "")
    m.required = ['Expression', 'Meaning']
    m.unique = ['Expression', 'Meaning']
    f = Field('Expression',
              _('A word or expression written in Hanzi.'))
    f.display['quiz']['fontSize'] = 72
    f.features.append("Reading source")
    m.addField(f)
    m.addField(Field('Meaning',
                          _(
        'A description in your native language, or Mandarin')))
    f = Field('Reading',
              _('Automatically generated by default.'))
    f.features.append("Reading destination")
    m.addField(f)
    m.addCard(CardModel("Production", _(
        "Actively test your recall by producing the target expression"),
                        "%(Meaning)s",
                        "%(Expression)s<br>%(Reading)s"))
    m.addCard(CardModel("Recognition", _(
        "Test your ability to recognize the target expression"),
                        "%(Expression)s",
                        "%(Reading)s<br>%(Meaning)s"))
    m.decorators.append("Mandarin")
    m.tags.append("Mandarin")
    return m
models['Mandarin'] = MandarinModel

