#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright: Damien Elmes <anki@ichi2.net>
# License: GNU GPL, version 2 or later; http://www.gnu.org/copyleft/gpl.html

"""\
Synchronisation
==============================

Support for keeping two decks synchronized. Both local syncing and syncing
over HTTP are supported.
"""
__docformat__ = 'restructuredtext'

import zlib
import re
import urllib, urllib2
import socket
import anki.json2 as json
import copy
import anki, anki.deck, anki.cards
from anki.errors import *
from anki.models import Model
from anki.fields import Field
from anki.cardmodels import CardModel
from anki.facts import Fact
from anki.cards import Card

# Protocol 2 code
##########################################################################

class SyncTools(object):

    def __init__(self, deck=None):
        self.deck = deck
        self.diffs = {}

    def setServer(self, server):
        self.server = server

    def modified(self):
        return self.deck.modified

    def unstuff(self, data):
        "Uncompress and convert to unicode."
        return json.read(unicode(zlib.decompress(data), "utf-8"))

    def stuff(self, data):
        "Convert into UTF-8 and compress."
        return zlib.compress(json.write(data).encode("utf-8"))

    # Control
    ##########################################################################

    def oneWay(self, op1, op2):
        # we need to copy the data when syncing locally
        if self.localTime < self.remoteTime:
            getattr(self, op1)(copy.deepcopy(getattr(self.server, op2)()))
        else:
            getattr(self.server, op1)(copy.deepcopy(getattr(self, op2)()))

    def sync(self):
        """Sync two local decks."""
        # check mod times
        if self.modified() == self.server.modified():
            return
        # record these mod times so that further operations don't corrupt them
        self.localTime = self.modified()
        self.remoteTime = self.server.modified()
        # get summaries
        self.sum = self.server.summary()
        # the order is importing here: models, facts, cards
        self.applyDiff("models", self.diffModels(self.sum['models']))
        self.applyDiff("facts", self.diffFacts(self.sum['facts']))
        self.applyDiff("cards", self.diffCards(self.sum['cards']))
        # finally, one-way data. ensure mod times are the same.
        self.oneWay("updateSched", "bundleSched")
        self.oneWay("updateFactsMeta", "bundleFactsMeta")
        self.oneWay("updateDeck", "bundleDeck")

    def applyDiff(self, type, diff):
        Type = type.capitalize()
        self.diffs[type] = diff
        # deletes on server and local
        if diff[1]:
            getattr(self.server, "delete" + Type)(diff[1])
        if diff[3]:
            getattr(self, "delete" + type.capitalize())(diff[3])
        # local & server updates: local first so we can rename
        if diff[2]:
            getattr(self, "update" + type.capitalize())(
                getattr(self.server, "get" + Type)(diff[2]))
        if diff[0]:
            getattr(self.server, "update" + Type)(
                getattr(self, "get" + Type)(diff[0]))

    # Updating from a fact diff
    ##########################################################################

    def getFacts(self, ids):
        return [self.bundleFact(f) for f in self.deck.facts if f.id in ids]

    def updateFacts(self, facts):
        for fact in facts:
            add = False
            local = self.deck.facts.ids.get(fact['id'])
            if not local:
                add = True
                local = Fact()
            local.update(fact['fields'])
            del fact["fields"]
            local.__dict__.update(fact)
            local.model = [m for m in self.deck.facts.models
                           if m.id == fact['model']][0]
            last = ([c for c in local.model.allcards
                     if c.id == fact['lastCard']])
            local.lastCard = last and last[0]
            local.parent = self.deck.facts
            # add cards for ids which exist locally. the others will be added
            # later
            local.cards = [self.deck.ids[id] for id in fact['cards']
                           if self.deck.ids.get(id, None) is not None]
            if add:
                self.deck.facts.addFact(local, validate=False)
            local.__dict__['modified'] = fact['modified']

    def deleteFacts(self, ids):
        for id in ids:
            # no need to remove cards, that will be done later
            self.deck.facts.remove(self.deck.facts.ids[id])

    # Updating from a card diff
    ##########################################################################

    def getCards(self, ids):
        return [self.bundleCard(c) for c in self.deck if c.id in ids]

    def updateCards(self, cards):
        for card in cards:
            add = False
            local = self.deck.ids.get(card['id'])
            if not local:
                add = True
                local = Card()
            local.__dict__.update(card)
            local.fact = [f for f in self.deck.facts if f.id == card['fact']][0]
            local.model = [c for c in local.fact.model.allcards
                           if c.id == card['model']][0]
            local.parent = self.deck
            if add:
                self.deck.addCard(local, len(self.deck))
                # update the fact now (we couldn't earlier)
                local.fact.cards.append(local)
            local.__dict__['modified'] = card['modified']

    def deleteCards(self, ids):
        for id in ids:
            self.deck.remove(self.deck.ids[id])

    # Updating from a model diff
    ##########################################################################

    def getModels(self, ids):
        return [self.bundleModel(m) for m in self.deck.facts.models
                if m.id in ids]

    def updateModels(self, models):
        for model in models:
            add = False
            local = [m for m in self.deck.facts.models
                     if m.id == model['id']]
            local = local and local[0]
            if not local:
                add = True
                local = Model()
            self.updateModel(local, model)
            if add:
                self.addModel(local)
            local.__dict__['modified'] = model['modified']

    def addModel(self, model):
        # add a model, renaming and existing, conflicting one
        local = self.deck.facts.models.byName(model.name)
        if local:
            # add the server one, without uniquifying
            self.deck.facts.models.append(model)
            # uniquify local
            self.deck.facts.models.makeModelUnique(local)
        else:
            self.deck.facts.models.addModel(model)

    def deleteModels(self, ids):
        for id in ids:
            model = [m for m in self.deck.facts.models
                     if m.id == id][0]
            # don't remove the associated cards
            self.deck.facts.removeModel(model, deleteFacts=False)

    def updateModel(self, model, meta):
        # preserve this attribute
        allcards = model.allcards
        model.__dict__.update(meta)
        model.allcards = allcards
        model.fields = [self.getField(f, model) for f in meta['fields']]
        for c in [self.getCardModel(f, model) for f in meta['allcards']]:
            self.updateCardModel(model.allcards, c)
        for n in range(len(model.cards)):
            model.cards[n] = [c for c in model.allcards
                              if c.id == model.cards[n]][0]

    def updateCardModel(self, models, model):
        for existing in models:
            if existing.id == model.id:
                existing.__dict__.update(model.__dict__)
                return
        # not found, create
        models.append(model)

    def getField(self, meta, parent):
        f = Field()
        f.__dict__.update(meta)
        f.parent = parent
        return f

    def getCardModel(self, meta, parent):
        c = CardModel()
        c.__dict__.update(meta)
        c.parent = parent
        return c

    # Bundling complex objects: render suitable for transport over JSON, etc
    ##########################################################################

    def bundleCard(self, card):
        "Return a card representation suitable for transport."
        c = dict(card.__dict__)
        c['fact'] = card.fact.id
        c['model'] = card.model.id
        del c['parent']
        return c

    def bundleFact(self, fact):
        "Return a fact representation suitable for transport."
        f = dict(fact.__dict__)
        f['fields'] = dict(fact)
        f['lastCard'] = fact.lastCard and fact.lastCard.id
        f['model'] = fact.model.id
        f['cards'] = [c.id for c in fact.cards]
        del f['parent']
        return f

    def bundleModel(self, model):
        "Return a model representation suitable for transport."
        m = dict(model.__dict__)
        m['fields'] = [self.bundleField(f) for f in model.fields]
        m['allcards'] = [self.bundleCardModel(c) for c in model.allcards]
        m['cards'] = [c.id for c in model.cards]
        del m['parent']
        return m

    def bundleModels(self):
        return [self.bundleModel(m) for m in self.deck.facts.models]

    def bundleField(self, field):
        "Return a field representation suitable for transport."
        f = dict(field.__dict__)
        del f['parent']
        return f

    def bundleCardModel(self, cardModel):
        "Return a card model representation suitable for transport."
        m = dict(cardModel.__dict__)
        del m['parent']
        return m

    # Bundling & updating simple structures
    ##########################################################################

    def bundleDeck(self):
        "Bundle up all one-way deck metadata."
        attrs = ("modified", "description")
        return dict([(k, getattr(self.deck, k)) for k in attrs])

    def bundleSched(self):
        "Bundle up all one-way sched metadata."
        attrs = ("maxScheduleTime", "hardInterval", "midInterval",
                 "easyInterval", "delay0", "delay1", "delay2",
                 "collapse", "globalStats", "dailyStats", "easeStats",
                 "highPriority", "medPriority", "lowPriority",
                 "suspendedTags")
        return dict([(k, getattr(self.deck.sched, k)) for k in attrs])

    def bundleFactsMeta(self):
        "Bundle up all one-way deck metadata."
        attrs = ("created", "modified")
        d = dict([(k, getattr(self.deck.facts, k)) for k in attrs])
        d['currentModel'] = self.deck.facts.models.current.id
        return d

    def updateSched(self, meta):
        self.deck.sched.__dict__.update(meta)

    def updateDeck(self, meta):
        self.deck.__dict__.update(meta)

    def updateFactsMeta(self, meta):
        id = meta['currentModel']
        del meta['currentModel']
        self.deck.facts.__dict__.update(meta)
        self.deck.facts.models.setModel(
            self.deck.facts.models.ids[id])

    # Summaries
    ##########################################################################

    def summary(self):
        "Generate a full summary of modtimes for two-way syncing."
        return {
            # covers cards
            "cards": self.cardSummary(),
            # covers facts (not models)
            "facts": self.factSummary(),
            # covers models, card models, fields
            "models": self.modelSummary(),
            }

    def cardSummary(self):
        return self.makeIdList(self.deck)

    def factSummary(self):
        return self.makeIdList(self.deck.facts)

    def modelSummary(self):
        return self.makeIdList(self.deck.facts.models)

    def makeIdList(self, attr):
        d = dict()
        d.update([(id, attr.ids[id] and attr.ids[id].modified or None)
                  for id in attr.ids.keys()])
        return {
            'modified': attr.modified,
            'ids': d,
            }

    # Diffing
    ##########################################################################

    def diffList(self, modified, localIds, summary):
        ltime = modified
        rtime = summary['modified']
        locallyEdited = []
        locallyDeleted = []
        remotelyEdited = []
        remotelyDeleted = []
        seen = {}
        # go through our local deck cache
        # build a hash of all ids, value being (localMod, remoteMod)
        ids = {}
        for (id, mod) in summary['ids'].items():
            ids[id] = [None, mod]
        for (id, obj) in localIds.items():
            if id in ids:
                ids[id][0] = obj and obj.modified or None
            else:
                ids[id] = [obj and obj.modified or None, None]
        # loop through the hash, determining differences
        for (id, (local, remote)) in ids.items():
            #print "test id=%d l=%s r=%s" % (id, `local`, `remote`)
            if local and remote:
                # changed/existing on both sides
                if local < remote:
                    remotelyEdited.append(id)
                elif local > remote:
                    locallyEdited.append(id)
            elif local and not remote:
                # deleted or missing on remote
                if id not in summary['ids']:
                    # missing
                    locallyEdited.append(id)
                else:
                    # deleted - compare against remote deck time
                    if local > rtime:
                        locallyEdited.append(id)
                    else:
                        remotelyDeleted.append(id)
            elif remote and not local:
                # deleted or missing on local
                if id not in localIds:
                    # missing
                    remotelyEdited.append(id)
                else:
                    # deleted - compare against local deck time
                    if remote > ltime:
                        remotelyEdited.append(id)
                    else:
                        locallyDeleted.append(id)
        return (locallyEdited, locallyDeleted,
                remotelyEdited, remotelyDeleted)

    def diffCards(self, summary):
        return self.diffList(self.deck.modified,
                             self.deck.ids,
                             summary)

    def diffFacts(self, summary):
        return self.diffList(self.deck.facts.modified,
                             self.deck.facts.ids,
                             summary)

    def diffModels(self, summary):
        return self.diffList(self.deck.facts.models.modified,
                             self.deck.facts.models.ids,
                             summary)


# Local syncing
##########################################################################

class SyncServer(SyncTools):
    pass

class SyncClient(SyncTools):
    pass

# HTTP proxy: act as a server and direct requests to the real server
##########################################################################

class HttpSyncServerProxy(SyncClient):

    def __init__(self, user, passwd):
        self.decks = None
        self.deckName = None
        self.username = user
        self.password = passwd
        self.syncURL="http://anki.ichi2.net/sync/"
        self.protocolVersion = 2
        self.setupMethods()

    def setupMethods(self):
        # one-way data methods
        for type in ("Sched", "Deck", "FactsMeta"):
            cmd = "bundle" + type
            setattr(self, cmd,
                    lambda cmd=cmd: self.unstuff(self.sendCmd(cmd)))
            cmd = "update" + type
            setattr(self, cmd,
                    lambda meta, cmd=cmd: \
                    self.sendCmd(cmd, data=self.stuff(meta)))
        # merging
        for type in ("Cards", "Facts", "Models"):
            cmd = "get" + type
            setattr(self, cmd,
                    lambda ids, cmd=cmd: \
                    self.unstuff(self.sendCmd(cmd, data=self.stuff(ids))))
            cmd = "update" + type
            setattr(self, cmd,
                    lambda obj, cmd=cmd: \
                    self.sendCmd(cmd, data=self.stuff(obj)))
            cmd = "delete" + type
            setattr(self, cmd,
                    lambda ids, cmd=cmd: \
                    self.sendCmd(cmd, data=self.stuff(ids)))

    def hasDeck(self, deckName):
        self.connect()
        return deckName.lower() in [d.lower() for d in self.decks.keys()]

    def availableDecks(self):
        self.connect()
        return self.decks.keys()

    def createDeck(self, deckName):
        self.sendCmd("createDeck", name=deckName)

    def summary(self):
        return self.unstuff(self.sendCmd("summary"))

    def modified(self):
        self.connect()
        if self.deckName not in self.decks:
            self.decks = None
            self.connect()
        return self.decks[self.deckName]

    def connect(self, clientVersion=""):
        "Check auth, protocol & grab deck list."
        if not self.decks:
            d = json.read(self.sendCmd("getDecks",
                                       lv=anki.version,
                                       pv=self.protocolVersion,
                                       cv=clientVersion))
            if d['status'] != "OK":
                raise SyncError(type="authFailed", status=d['status'])
            self.decks = d['decks']

    def sendCmd(self, action, **args):
        data = {"d": self.deckName,
                "p": self.password,
                "u": self.username}
        data.update(args)
        data = urllib.urlencode(data)
        try:
            f = urllib2.urlopen(self.syncURL + action, data)
        except (urllib2.URLError, socket.error, socket.timeout):
            raise SyncError(type="noResponse")
        ret = f.read()
        if not ret:
            raise SyncError(type="noResponse")
        return ret

    def saveDeck(self):
        self.sendCmd("saveDeck")

# HTTP server: respond to proxy requests and return data
##########################################################################

class HttpSyncServer(SyncServer):

    def __init__(self):
        SyncServer.__init__(self)
        self.protocolVersion = 2
        self.decks = {}
        self.setupMethods()

    def ok(self, *args):
        "Evalate args then return OK."
        return "OK"

    def setupMethods(self):
        # one-way data methods
        for type in ("Sched", "Deck", "FactsMeta"):
            cmd = "bundle" + type
            setattr(self, cmd,
                    lambda cmd=cmd: \
                    self.stuff(getattr(SyncServer, cmd)(self)))
            cmd = "update" + type
            setattr(self, cmd,
                    lambda meta, cmd=cmd: \
                    self.ok(getattr(SyncServer, cmd)(
                self, self.unstuff(meta))))
        # merging
        for type in ("Cards", "Facts", "Models"):
            cmd = "get" + type
            setattr(self, cmd,
                    lambda ids, cmd=cmd: \
                    self.stuff(getattr(SyncServer, cmd)(
                self, self.unstuff(ids))))
            cmd = "update" + type
            setattr(self, cmd,
                    lambda obj, cmd=cmd: \
                    self.ok(getattr(SyncServer, cmd)(
                self, self.unstuff(obj))))
            cmd = "delete" + type
            setattr(self, cmd,
                    lambda ids, cmd=cmd: \
                    self.ok(getattr(SyncServer, cmd)(
                self, self.unstuff(ids))))

    def getDecks(self, lv, pv, cv):
        return json.write({
            "status": "OK",
            "decks": self.decks,
            })

    def summary(self):
        return self.stuff(SyncServer.summary(self))
