# coding: utf-8

import nose, os
import anki
from anki import stdmodels
from anki.errors import *
from anki.facts import *
from anki.models import *
from anki.fields import *

deck = None
ds = None
testDir = os.path.dirname(__file__)

def setup1():
    global deck, ds
    deck = anki.Controller.Deck()
    ds = anki.Controller.DeckStorage()
    deck.facts.models.current.cards = deck.facts.models.current.allcards

def assertException(exception, func):
    found = False
    try:
        func()
    except exception:
        found = True
    assert found

#
##########################################################################

def addAuto(n=""):
    f = Fact()
    for field in deck.facts.models.current.fields:
        f[field.name] = field.name + n
    deck.addFact(f)

def addBunch():
    for i in range(10):
        addAuto(str(i))

@nose.with_setup(setup1)
def test_importoldv7():
    file = unicode(os.path.join(testDir, "importing/test-v8.fc"))
    # we shouldn't be able to load it normally
    assertException(DeckWrongFormatError,
                    lambda: ds.load(file))
    # but we can open it directly for importing later
    d = ds.fromFile(file)
    assert len(d) == 5
    c = d[0]
    assert c.question and c.answer
    assert c.nextTime and getattr(c, 'interval', 1) == 0

@nose.with_setup(setup1)
def test_addcards2():
    f = Fact()
    f['Front'] = "foo"; f['Back'] = "bar"
    deck.addFact(f)
    assert len(deck) == 2

@nose.with_setup(setup1)
def test_addcards1():
    del deck.facts.models.current.cards[0]
    f = Fact()
    f['Front'] = "foo"; f['Back'] = "bar"
    deck.addFact(f)
    assert len(deck) == 1

@nose.with_setup(setup1)
def test_addcardsjp():
    deck.facts.models.setModel(stdmodels.byName('Japanese'))
    f = Fact()
    f['Expression'] = "foo"; f['Reading'] = "bar"; f['Meaning'] = "baz"
    deck.addFact(f)
    assert len(deck) == 2

@nose.with_setup(setup1)
def test_addcardsauto():
    addAuto()
    assert len(deck) == 2

@nose.with_setup(setup1)
def test_sched1():
    addBunch()
    card = deck.sched.getCard()
    assert card
    assert deck.sched.cardIsNew(card)
    assert deck.sched.cardIsYoung(card)
    assert deck.sched.cardIsInInitialState(card)

@nose.with_setup(setup1)
def test_fieldsandmodels():
    addBunch()
    m = stdmodels.byName('Other')
    m.cards = m.allcards
    f = Fact()
    f['Front'] = "foox"
    f['Back'] = "fooy"
    deck.facts.models.setModel(m)
    deck.addFact(f)
    stdm = deck.facts.models[0]
    assert len(deck) == 22
    assert deck.modelUseCount(m) == 2
    assert deck.modelUseCount(stdm) == 20
    # removing fields
    assert deck.fieldUseCount(m, 'Front') == 2
    assert deck.fieldUseCount(stdm, 'Front') == 20
    deck.facts.removeField(stdm,
                           stdm.fieldByName('Front'))
    assert deck.fieldUseCount(m, 'Front') == 2
    assert deck.fieldUseCount(stdm, 'Front') == 0
    assert deck.fieldUseCount(m, 'Back') == 2
    assert deck.fieldUseCount(stdm, 'Back') == 20
    deck.facts.removeField(stdm,
                           stdm.fieldByName('Back'))
    assert deck.fieldUseCount(m, 'Back') == 2
    assert deck.fieldUseCount(stdm, 'Back') == 0
    # adding fields
    deck.facts.addField(stdm, Field('Back'))
    # even though we've added it to 10 cards, it's currently empty and thus
    # doesn't register
    assert deck.fieldUseCount(stdm, 'Back') == 0
    # but they should register if we put something in them
    for fact in deck.facts:
        if fact.model == stdm:
            fact['Back'] = "hi"
    assert deck.fieldUseCount(stdm, 'Back') == 20
    # now let's rename 'Back' to something else
    deck.facts.renameField(stdm,
                           stdm.fieldByName('Back'),
                           'Sideways')
    assert deck.fieldUseCount(stdm, 'Back') == 0
    assert deck.fieldUseCount(stdm, 'Sideways') == 20
    # removing a model
    assert len(deck) == 22
    assert deck.modelUseCount(m) == 2
    deck.removeModel(m)
    assert len(deck) == 20
    assert deck.modelUseCount(m) == 0
    assert deck.cardModelUseCount(stdm.allcards[0]) == 10

@nose.with_setup(setup1)
def test_removecard():
    f = Fact()
    f['Front'] = "x"
    f['Back'] = "y"
    assert len(deck) == 0
    cards = deck.addFact(f)
    assert len(deck) == 2
    assert len(deck.facts) == 1
    deck.remove(cards[0])
    assert len(deck) == 1
    assert len(deck.facts) == 1
    deck.remove(cards[1])
    assert len(deck) == 0
    assert len(deck.facts) == 0

@nose.with_setup(setup1)
def test_alltags():
    deck.facts.models.current.tags.append("foo")
    f = Fact()
    f['Front'] = "x"
    f['Back'] = "y"
    f.tags.append("bar")
    deck.addFact(f)
    t = deck.allTags()
    assert t['model'] == ['foo']
    assert t['fact'] == ['bar']
    assert len(t['card']) == 0
    assert len(t['cardModel']) == 2
    assert 'Front to back' in t['cardModel']
    assert 'Back to front' in t['cardModel']
    assert len(t['all']) == 4

# modification tracking
##########################################################################

def assertChanged(func):
    # stupid windows
    deck.lastLoaded = time.time() - 1
    deck.modified = deck.lastLoaded
    assert not deck.modifiedSinceSave()
    func()
    assert deck.modifiedSinceSave()

@nose.with_setup(setup1)
def test_modtime():
    f = Fact()
    f['Front'] = "foo"
    f['Back'] = "bar"
    assertChanged(lambda f=f: deck.addFact(f))
    assertChanged(lambda f=f: f.__setitem__("Front", "new"))
    assertChanged(lambda f=f: setattr(f, "interval", 5))
    assertChanged(lambda: deck.removeField(
        deck.facts.models.current,
        deck.facts.models.current.fields[0]))
    # empty field
    f = Field('abc')
    deck.facts.addField(deck.facts.models.current, f)
    assertChanged(lambda f=f: deck.removeField(
        deck.facts.models.current, f))
