# coding: utf-8

import nose
from anki.errors import *
from anki.facts import *
from anki.models import *
from anki import stdmodels

facts = None
model = None
fact = None

def setup1():
    global facts, model, fact
    facts = Facts()
    model = stdmodels.byName('Other')
    facts.models.setModel(model)
    fact = Fact()
    fact['Front'] = "foo"; fact['Back'] = "bar";
    fact.tags.append("test")

def assertException(exception, func):
    found = False
    try:
        func()
    except exception:
        found = True
    assert found

# Facts
##########################################################################

@nose.with_setup(setup1)
def test_facts_ids():
    global fact
    facts.addFact(fact)
    assert fact.id == hash("barfoo")
    fact = Fact()
    fact['Front'] = "baz"; fact['Back'] = "quux"
    facts.addFact(fact)
    assert fact.id == hash("bazquux")

# Facts + models
##########################################################################

@nose.with_setup(setup1)
def test_fieldsvalid():
    model.ensureValid(fact)
    fact['Front'] = " "
    assertException(FactInvalidError,
                    lambda: model.ensureValid(fact))

@nose.with_setup(setup1)
def test_isunique():
    facts.addFact(fact)
    f = Fact()
    f['Front'] = "a"; f['Back'] = "b"
    facts.models.current.ensureUnique(f, facts)
    facts.addFact(f)
    f = Fact()
    f['Front'] = "foof"; f['Back'] = "barf"
    facts.models.current.ensureUnique(f, facts)
    f = Fact()
    f['Front'] = "foo"; f['Back'] = "barf"
    assertException(FactInvalidError,
                    lambda: model.ensureUnique(f, facts))
    facts.models.current.ensureValid(fact)

@nose.with_setup(setup1)
def test_multimodels():
    global model, fact
    facts.addFact(fact)
    m2 = stdmodels.byName("Other")
    facts.models.setModel(m2)
    fact = Fact()
    fact['Front'] = "qqq"; fact['Back'] = "ppp"
    facts.addFact(fact)
    assert len(facts.models) == 2
    assert facts.models.current.name == m2.name

@nose.with_setup(setup1)
def test_cardfmt():
    facts.addFact(fact)
    card = facts.models.current.cards[0].makeCard(fact)
    assert card.question == "foo" and card.answer == "bar"
    # change the rendering for our current method
    facts.models.current.cards[0].qformat = "qux"
    assert card.question == "qux" and card.answer == "bar"
    m2 = stdmodels.byName("Other")
    facts.models.setModel(m2)
    assert len(facts.models) == 2
    # ensure the cards generated with the old method keep the old format, but
    # cards generated with the new method are back to default
    card2 = facts.models.current.cards[0].makeCard(fact)
    # ensure that we can add with the old format too
    card3 = facts.models[0].cards[0].makeCard(fact)
    assert card.question == "qux" and card.answer == "bar"
    assert card2.question == "foo" and card2.answer == "bar"
    assert card3.question == "qux" and card2.answer == "bar"
    # allowing multiple methods whose names don't differ is confusing
    assert facts.models[0].name != facts.models[1].name

@nose.with_setup(setup1)
def test_byField():
    facts.addFact(fact)
    f = Fact(); f['Front'] = "foof"; f['Back'] = "barf"
    f.tags.append("test2")
    facts.addFact(f)
    f = Fact(); f['Front'] = "ofoo"; f['Back'] = "far";
    f.tags.append("test2")
    f.tags.append("test3")
    facts.addFact(f)
    f = Fact(); f['Front'] = "bar"; f['Back'] = "foo";
    f.tags.append("best")
    facts.addFact(f)
    assert len(facts) == 4
    res = facts.byField("Front", "foo")
    assert len(res) == 1
    assert res[0]['Front'] == "foo"
    res = facts.byField("Back", "bar")
    assert len(res) == 1
    assert res[0]['Back'] == "bar"
    # we should be able to get tags, too
    res = facts.byTag("test")
    assert len(res) == 1
    # and substrings
    res = facts.byField("Front", "foo", True)
    assert len(res) == 3
    # if we change the model and search for a new field, it should ignore
    # different model types
    m2 = stdmodels.byName("Japanese")
    facts.models.setModel(m2)
    f = Fact()
    f['Expression'] = "foo"; f['Meaning'] = "bar"; f['Reading'] = ""
    facts.addFact(f)
    assert len(facts.byField("Expression", "foo")) == 1

@nose.with_setup(setup1)
def test_cardgen():
    global fact
    facts.addFact(fact)
    assert len(facts.models.current.cards) == 1
    card = facts.models.current.cards[0].makeCard(fact)
    assert card.question == "foo" and card.answer == "bar"
    m2 = stdmodels.byName("Japanese")
    facts.models.setModel(m2)
    fact = Fact()
    fact['Expression'] = "a"
    fact['Reading'] = "b"
    fact['Meaning'] = "c"
    facts.addFact(fact)
    cards = [cmodel.makeCard(fact) for cmodel in facts.models.current.cards]
    card = [c for c in cards if c.model.name == "Recognition"][0]
    assert card.question == "a" and card.answer == "b<br>c"
    card = [c for c in cards if c.model.name == "Production"][0]
    assert card.question == "c" and card.answer == "a<br>b"

@nose.with_setup(setup1)
def test_addremove():
    facts.addFact(fact)
    # same object
    # FIXME: for now, we don't die here
    #assertException(FactInvalidError,
    #                lambda: facts.addFact(fact))
    f = Fact()
    f['Front'] = "foo"; f['Back'] = "ba2r"
    # similar object
    assertException(FactInvalidError,
                    lambda: facts.addFact(f))
    # remove earlier object
    assert len(facts) == 1
    id = facts[0].id
    assert id is not None
    del facts[0]
    # add it again. it should be given the same id
    fact['id'] = None
    facts.addFact(fact)
    assert fact.id == id
    # remove it again with remove()
    facts.remove(fact)
    assert len(facts) == 0

@nose.with_setup(setup1)
def test_removemodel():
    facts.addFact(fact)
    # add japanese in the middle
    jmodel = stdmodels.byName("Japanese")
    facts.models.setModel(jmodel)
    f2 = Fact()
    f2['Expression'] = "a"
    f2['Reading'] = "b"
    f2['Meaning'] = "c"
    facts.addFact(f2)
    # and another 2 simple facts for good measure
    f2 = Fact()
    f2['Front'] = "foo1"
    f2['Back'] = "foo2"
    model2 = stdmodels.byName("Other")
    facts.models.setModel(model2)
    facts.addFact(f2)
    f2 = Fact()
    f2['Front'] = "foo3"
    f2['Back'] = "foo4"
    facts.addFact(f2)
    assert len(facts.models) == 3
    assert facts.modelUseCount(model) == 1
    assert facts.modelUseCount(jmodel) == 1
    assert facts.modelUseCount(model2) == 2
    facts.removeModel(model2)
    assert facts.modelUseCount(model2) == 0
    assert len(facts) == 2
    facts.removeModel(model)
    assert facts.modelUseCount(model) == 0
    assert len(facts) == 1
