# coding: utf-8

import anki
from anki.sync import *
from anki.models import *
from anki.facts import *
from anki.fields import *
from anki import stdmodels
import nose

# Local tests
##########################################################################

deck1=None
deck2=None
client=None
server=None

def setup_local():
    global deck1, deck2, client, server
    deck1 = anki.Controller.Deck()
    deck1.facts.models.current.cards = deck1.facts.models.current.allcards
    f = Fact()
    f['Front'] = "foo"; f['Back'] = "bar"
    deck1.addFact(f)
    # stupid windows
    time.sleep(0.01)
    deck2 = anki.Controller.Deck()
    deck2.facts.models.current.cards = deck2.facts.models.current.allcards
    f = Fact()
    f['Front'] = "baz"; f['Back'] = "qux"
    deck2.addFact(f)
    client = SyncClient(deck1)
    server = SyncServer(deck2)
    client.setServer(server)

def teardown():
    pass

@nose.with_setup(setup_local, teardown)
def test_localsync_diffing1():
    assert len(deck1) == 2
    assert len(deck2) == 2
    result = client.diffCards(server.summary()['cards'])
    assert len(result[0]) == 2
    assert len(result[1]) == 0
    assert len(result[2]) == 2
    assert len(result[3]) == 0
    result = client.diffFacts(server.summary()['facts'])
    assert len(result[0]) == 1
    assert len(result[1]) == 0
    assert len(result[2]) == 1
    assert len(result[3]) == 0

@nose.with_setup(setup_local, teardown)
def test_localsync_oneway():
    # deck two was modified last
    assert deck2.modified > deck1.modified
    d2mod = deck2.modified
    client.sync()
    assert deck1.modified == deck2.modified
    # make sure we're not copying actual instances
    assert id(deck1.sched.lowPriority) != id(deck2.sched.lowPriority)
    # ensure values are being synced
    deck1.sched.lowPriority.append("foo")
    deck1.updatePriorities()
    time.sleep(0.01)
    deck1.setModified()
    client.sync()
    assert "foo" in deck2.sched.lowPriority
    deck2.description = "newname"
    time.sleep(0.01)
    deck2.setModified()
    client.sync()
    assert deck1.description == "newname"
    # the most recent change should take precedence
    deck1.description = "foo"
    time.sleep(0.01)
    deck1.setModified()
    deck2.description = "bar"
    time.sleep(0.01)
    deck2.setModified()
    client.sync()
    assert deck1.description == "bar"

@nose.with_setup(setup_local, teardown)
def test_localsync_models():
    assert len(deck1) == 2
    assert len(deck2) == 2
    client.sync()
    assert len(deck1.facts.models) == 2
    assert len(deck2.facts.models) == 2
    mod = Model("test")
    time.sleep(0.01)
    deck1.facts.models.setModel(mod)
    assert len(deck1.facts.models) == 3
    assert len(deck2.facts.models) == 2
    assert mod not in deck2.facts.models
    # sync should copy the model
    client.sync()
    assert len(deck1.facts.models) == 3
    assert len(deck2.facts.models) == 3
    # and properties
    mod.name = "test2"
    time.sleep(0.01)
    mod.setModified()
    assert [m for m in deck1.facts.models if m.name == "test2"]
    assert not [m for m in deck2.facts.models if m.name == "test2"]
    client.sync()
    assert [m for m in deck1.facts.models if m.name == "test2"]
    assert [m for m in deck2.facts.models if m.name == "test2"]

@nose.with_setup(setup_local, teardown)
def test_localsync_adddeleteadd_model():
    client.sync()
    m = stdmodels.byName("Japanese")
    m.name = "new"
    time.sleep(0.01)
    deck1.facts.models.setModel(m)
    assert len(deck1.facts.models) == 3
    assert len(deck2.facts.models) == 2
    client.sync()
    assert len(deck1.facts.models) == 3
    assert len(deck2.facts.models) == 3
    time.sleep(0.01)
    deck1.removeModel(m)
    assert len(deck1.facts.models) == 2
    assert len(deck2.facts.models) == 3
    client.sync()
    assert len(deck1.facts.models) == 2
    assert len(deck2.facts.models) == 2
    time.sleep(0.01)
    deck1.facts.models.setModel(m)
    assert len(deck1.facts.models) == 3
    assert len(deck2.facts.models) == 2
    client.sync()
    assert len(deck1.facts.models) == 3
    assert len(deck2.facts.models) == 3

@nose.with_setup(setup_local, teardown)
def test_localsync_facts():
    client.sync()
    result = client.diffFacts(server.summary()['facts'])
    assert len(result[0]) == 0
    assert len(result[1]) == 0
    assert len(result[2]) == 0
    assert len(result[3]) == 0

@nose.with_setup(setup_local, teardown)
def test_localsync_cards():
    client.sync()
    result = client.diffCards(server.summary()['cards'])
    assert len(result[0]) == 0
    assert len(result[1]) == 0
    assert len(result[2]) == 0
    assert len(result[3]) == 0

# make sure we can add/delete and read the same fact
@nose.with_setup(setup_local, teardown)
def test_localsync_deleteandadd_cards1():
    client.sync()
    # add a new question
    time.sleep(0.01)
    f = Fact()
    f['Front'] = "a"; f['Back'] = "b"
    cards = deck1.addFact(f)
    card = cards[0]
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 2
    assert len(diff[1]) == 0
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    # delete it
    time.sleep(0.01)
    deck1.remove(card)
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 0
    assert len(diff[1]) == 1
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    # and add again
    time.sleep(0.01)
    cards = [card.model.makeCard(card.fact), cards[1]]
    deck1.addCard(cards[0], 0)
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 1
    assert len(diff[1]) == 0
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    # delete both
    time.sleep(0.01)
    for card in cards:
        deck1.remove(card)
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 0
    assert len(diff[1]) == 2
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    assert len(deck1) == 4
    assert len(deck2) == 4

@nose.with_setup(setup_local, teardown)
def test_localsync_threeway():
    # deck1 (client) <-> deck2 (server) <-> deck3 (client)
    deck3 = anki.Controller.Deck()
    client2 = SyncClient(deck3)
    server2 = SyncServer(deck2)
    client2.setServer(server2)
    client.sync()
    client2.sync()
    # add a new question
    time.sleep(0.01)
    f = Fact()
    f['Front'] = "a"; f['Back'] = "b"
    cards = deck1.addFact(f)
    card = cards[0]
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 2
    assert len(diff[1]) == 0
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    assert len(deck1) == 6
    assert len(deck2) == 6
    # check it propagates from server to deck3
    client2.sync()
    assert len(deck3) == 6
    # delete a card on deck1
    time.sleep(0.01)
    deck1.remove(card)
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 0
    assert len(diff[1]) == 1
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    assert len(deck1) == 5
    assert len(deck2) == 5
    # make sure the delete is now propagated from the server to deck3
    diff = client2.diffCards(server2.summary()['cards'])
    assert len(diff[0]) == 0
    assert len(diff[1]) == 0
    assert len(diff[2]) == 0
    assert len(diff[3]) == 1
    client2.sync()
    assert len(deck3) == 5

@nose.with_setup(setup_local, teardown)
def test_localsync_deleteandadd_facts():
    client.sync()
    # add a new fact
    time.sleep(0.01)
    f = Fact()
    f['Front'] = "a"; f['Back'] = "b"
    cards = deck1.addFact(f)
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 2
    assert len(diff[1]) == 0
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    # delete the cards
    time.sleep(0.01)
    for card in cards:
        deck1.remove(card)
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 0
    assert len(diff[1]) == 2
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    # and add again
    time.sleep(0.01)
    f = Fact()
    f['Front'] = "a"; f['Back'] = "b"
    cards = deck1.addFact(f)
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 2
    assert len(diff[1]) == 0
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    # delete both again
    time.sleep(0.01)
    for card in cards:
        deck1.remove(card)
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 0
    assert len(diff[1]) == 2
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    assert len(deck1) == 4
    assert len(deck2) == 4
    # and add
    time.sleep(0.01)
    f = Fact()
    f['Front'] = "a"; f['Back'] = "b"
    cards = deck1.addFact(f)
    diff = client.diffCards(server.summary()['cards'])
    assert len(diff[0]) == 2
    assert len(diff[1]) == 0
    assert len(diff[2]) == 0
    assert len(diff[3]) == 0
    client.sync()
    # delete
    c = deck1.index(cards[1])
    del deck1[c]
    client.sync()

@nose.with_setup(setup_local, teardown)
def test_localsync_cardmodels():
    deck1.facts.models.current.cards = deck1.facts.models.current.allcards
    client.sync()
    deck1.facts.models.current.allcards[0].aformat = "foo"
    time.sleep(0.01)
    deck1.facts.models.current.allcards[0].setModified()
    assert deck2.facts.models.current.allcards[0].aformat != "foo"
    deck2.description = "2"
    client.sync()
    assert deck2.facts.models.current.allcards[0].aformat == "foo"

@nose.with_setup(setup_local, teardown)
def test_localsync_cardmodels_reinit():
    global deck1
    client.sync()
    assert deck1.facts.models.current.id == deck2.facts.models.current.id
    # reinit the first deck
    deck1 = anki.Controller.Deck()
    client.deck = deck1
    client.sync()
    assert deck1.facts.models.current.id == deck2.facts.models.current.id

@nose.with_setup(setup_local, teardown)
def test_localsync_currentmodel():
    assert deck1.facts.models.current != deck2.facts.models.current
    client.sync()
    assert deck1.facts.models.current == deck2.facts.models.current

@nose.with_setup(setup_local, teardown)
def test_localsync_displayproperties():
    e2 = deck2.facts.models.current.fieldByName("Front")
    e2.display['quiz']['fontSize'] = 300
    e1 = deck1.facts.models.current.fieldByName("Front")
    assert e1.display['quiz']['fontSize'] != 300
    client.sync()
    e1 = deck1.facts.models.current.fieldByName("Front")
    assert e1.display['quiz']['fontSize'] == 300

@nose.with_setup(setup_local, teardown)
def test_localsync_changeanddelete():
    client.sync()
    time.sleep(0.01)
    deck2[0].setModified()
    time.sleep(0.01)
    deck1.remove(deck2[0])
    client.sync()

# Remote tests
##########################################################################

# a replacement sendCmd which just calls our server
def sendCmd(action, *args, **kargs):
    #print action, kargs
    return apply(getattr(server, action), tuple(args) +
                 tuple(kargs.values()))

def setup_remote():
    setup_local()
    global client, server
    proxy = HttpSyncServerProxy("test", "foo")
    client = SyncClient(deck1)
    client.setServer(proxy)
    proxy.deckName = "test"
    proxy.sendCmd = sendCmd
    server = HttpSyncServer()
    server.deck = deck2
    server.decks = {"test": deck2.modified}

@nose.with_setup(setup_remote, teardown)
def test_remotesync_fromserver():
    # deck two was modified last
    assert deck2.modified > deck1.modified
    client.sync()

@nose.with_setup(setup_remote, teardown)
def test_remotesync_toserver():
    deck1.modified = time.time()
    client.sync()

def test_remotesync_specialchars():
    deck1[0].fact['Expression'] = "'\\\"::}{]\n\n[()"
    client.sync()

# Json2
##########################################################################

from anki.json2 import *
import types

def test_json2():
    # standard json
    assert not isinstance(read(write({"foo": 5})).keys()[0], types.UnicodeType)
    assert not isinstance(read(write({u"foo": 5})).keys()[0], types.UnicodeType)
    assert isinstance(read(write({u"食べる": 5})).keys()[0], types.UnicodeType)
    # stuffing/unstuffing
    assert not isinstance(client.unstuff(
        client.stuff({"foo": 5})).keys()[0], types.UnicodeType)
    # unicode will be converted to standard text if it's ascii
    assert not isinstance(client.unstuff(
        client.stuff({u"foo": 5})).keys()[0], types.UnicodeType)
    # but not if it contains special characters
    assert isinstance(client.unstuff(
        client.stuff({u"食べる": 5})).keys()[0], types.UnicodeType)
